from __future__ import annotations

import math
import os
from copy import copy
from dataclasses import dataclass
from typing import Any, Callable, Dict, List, Optional, Sequence, Set, Tuple, TypeVar, Union
from warnings import warn

import torch

import rust_circuit.optional as op
from rust_circuit.algebric_rewrite import check_permutation
from rust_circuit.py_utils import assert_never, make_index_at
from rust_circuit.ui.very_named_tensor import VeryNamedTensor, ViewSpec, ViewSpecIdx

from .._rust import Circuit, Index, Shape, TorchAxisIndex

EvalCallback = Callable[[Circuit, ViewSpec], Union[Tuple[torch.Tensor, Set[int], Set[int]], torch.Tensor]]
""" returns evaluated tensor, handled view items, removed dim items """

T = TypeVar("T")


@dataclass
class CircuitsVeryNamedTensor:
    def __init__(
        self,
        circuit: Circuit,
        eval_callback: EvalCallback,
        dim_names: Sequence[str],
        dim_types: Sequence[str],
        dim_idx_names: Sequence[Sequence[str]],
        units: str = "units",
        title: str = "untitled_tensor",
        aux_info: Any = {},
        selected_options=None,
        default_axis_view: ViewSpecIdx = "axis",
        default_view_spec: Optional[ViewSpec] = None,  # overrides axis
        extra_shape: Shape = (),
        skip_last_dims: int = 0,
        index_device: Union[str, torch.device] = "cpu",
        caching: bool = True,
        caching_numel: int = 1024 ** 3 // 16,
        caching_perm: Optional[List[Union[int, str]]] = None,
        custom_can_be_canon_view: Optional[Callable[[Tuple[int, ...], ViewSpec, ViewSpec], bool]] = None,
        always_allow_cache_on_numel: bool = True,
        prefer_index_axis_caching: bool = True,
    ):
        self.circuit = circuit

        self.eval_callback = eval_callback

        self.dim_names = dim_names
        self.dim_types = dim_types
        self.dim_idx_names = [[str(i) for i in x] for x in dim_idx_names]
        self.units: str = units
        self.title = title

        self.aux_info = aux_info
        self.selected_options = selected_options
        via_axis: ViewSpec = [default_axis_view] * len(self.dim_names)
        self.default_view_spec: ViewSpec = op.unwrap_or(default_view_spec, via_axis)

        self.extra_shape = extra_shape
        self.skip_last_dims = skip_last_dims
        self.index_device = index_device

        # TODO: should caches be on cpu???
        self.caching = caching
        self.caching_numel = caching_numel
        self.caching_perm = caching_perm
        self.custom_can_be_canon_view = custom_can_be_canon_view
        self.prefer_index_axis_caching = prefer_index_axis_caching
        self.always_allow_cache_on_numel = always_allow_cache_on_numel

        self.cache: Dict[Tuple[Union[Tuple[ViewSpecIdx, ...], ViewSpecIdx], ...], VeryNamedTensor] = {}

        _ = self.caching_perm_val  # call to trigger asserts
        assert self.skip_last_dims >= 0, self.skip_last_dims
        assert self.skip_last_dims <= self.circuit.rank
        self.avoid_count = len(self.extra_shape) + self.skip_last_dims
        self.non_avoid_count = self.ndim - self.avoid_count

        assert len(dim_names) == len(
            self.shape
        ), f"{len(dim_names)} dim_names given ({dim_names}), but tensor is of rank {len(self.shape)} ({self.shape})"
        assert len(dim_types) == len(
            self.shape
        ), f"{len(dim_types)} dim_types given ({dim_types}), but tensor is of rank {len(self.shape)} ({self.shape})"
        idx_shape = tuple([len(g) for g in self.dim_idx_names])
        assert (
            idx_shape == self.shape
        ), f"dim_idx_names of shape {idx_shape} does not match tensor of shape {self.shape}"

    @property
    def caching_perm_val(self):
        if self.caching_perm is None:
            out = list(range(len(self.dim_names)))
        else:
            out = [(i if isinstance(i, int) else self.dim_names.index(i)) for i in self.caching_perm]
        check_permutation(out, len(self.dim_names))

        return out

    @property
    def shape(self):
        return self.circuit.shape + self.extra_shape

    @property
    def ndim(self):
        return len(self.shape)

    @staticmethod
    def try_idx_view(
        view_el: Union[ViewSpecIdx, List[ViewSpecIdx]], index_device: Union[str, torch.device] = "cpu"
    ) -> Optional[Union[int, torch.Tensor]]:
        if isinstance(view_el, int):
            return view_el
        elif isinstance(view_el, list):
            all_idxs: List[int] = []
            for i in view_el:
                if not isinstance(i, int):
                    return None
                all_idxs.append(i)
            return torch.tensor(all_idxs, device=index_device)

        return None

    @classmethod
    def apply_view_to_circ(
        cls,
        circ: Circuit,
        view_el: Union[ViewSpecIdx, List[ViewSpecIdx]],
        dim: int,
        index_device: Union[str, torch.device] = "cpu",
    ) -> Optional[Tuple[Circuit, bool]]:
        """returns new circuit and whether or not the dim was removed"""

        if (idx := cls.try_idx_view(view_el, index_device=index_device)) is not None:
            return Index(circ, make_index_at(idx, dim)), len(cls.get_view_shape_handled(view_el, circ.shape[dim])) == 0
        elif view_el == "mean":
            return circ.mean(axis=dim), True
        elif view_el == "sum":
            return circ.sum(axis=dim), True

        return None

    @classmethod
    def get_view_shape_handled(cls, view_el: Union[ViewSpecIdx, List[ViewSpecIdx]], size: int):
        return cls.get_view_shape_handled_status(view_el, size)[0]

    @classmethod
    def get_view_shape_handled_status(cls, view_el: Union[ViewSpecIdx, List[ViewSpecIdx]], size: int):
        if (idx := cls.try_idx_view(view_el)) is not None:
            if isinstance(idx, int):
                out_shape = ()
            elif isinstance(idx, torch.Tensor):
                out_shape = (len(idx),)
            else:
                assert_never(idx)

            assert len(out_shape) in [0, 1]
            return out_shape, True
        elif view_el == "mean":
            return (), True
        elif view_el == "sum":
            return (), True
        return (size,), False

    @staticmethod
    def apply_idx_to_list(idx: Union[slice, torch.Tensor], lst: List[T]) -> List[T]:
        if isinstance(idx, slice):
            return lst[idx]
        assert idx.ndim == 1
        return [lst[int(i)] for i in idx]

    def default_can_be_canon_view(self, sizes: Tuple[int, ...]):
        return math.prod(sizes) < self.caching_numel

    def can_be_canon_view(self, sizes: Tuple[int, ...], view: ViewSpec, orig_view: ViewSpec) -> bool:
        assert len(sizes) == len(view)
        if self.custom_can_be_canon_view is not None:
            return self.custom_can_be_canon_view(sizes, view, orig_view)
        else:
            return self.default_can_be_canon_view(sizes)

    @classmethod
    def strip_leftmost_index(cls, view: ViewSpec):
        view = copy(view)
        idx = min((i for i, v in enumerate(view) if cls.try_idx_view(v) is not None), default=None)
        if idx is None:
            return None
        view[idx] = "axis"
        return view

    def get_canonical_cache_suffix_view(
        self,
        perm_view: ViewSpec,
        perm_shape: Tuple[int, ...],
        perm_back: List[int],
        orig_view: ViewSpec,
        is_base: bool = True,
    ) -> Optional[ViewSpec]:
        # this caching logic is kinda gross and sad... : /
        index_stripped = self.strip_leftmost_index(perm_view)
        if self.prefer_index_axis_caching and index_stripped is not None:
            if (
                rec_out := self.get_canonical_cache_suffix_view(
                    index_stripped, perm_shape, perm_back, orig_view, is_base=False
                )
            ) is not None:
                return rec_out

        vals = [self.get_view_shape_handled_status(view_el, size) for view_el, size in zip(perm_view, perm_shape)]
        sizes = tuple(math.prod(s) for s, _ in vals)
        for i in reversed(range(self.ndim + 1)):
            target_view_start: ViewSpec = ["axis"] * i
            target_view = target_view_start + [("sum" if x == "mean" else x) for x in perm_view[i:]]
            new_p_shape = perm_shape[:i] + sizes[i:]
            un_perm_shape = tuple(new_p_shape[p] for p in perm_back)
            un_perm_view = [target_view[p] for p in perm_back]
            if self.can_be_canon_view(un_perm_shape, un_perm_view, orig_view) or (
                i == 0
                and is_base
                and self.default_can_be_canon_view(un_perm_shape)
                and self.always_allow_cache_on_numel
            ):
                return target_view

        return None

    def get_canonical_cache_suffix_view_perm(self, view: ViewSpec) -> Optional[ViewSpec]:
        perm_view = [view[p] for p in self.caching_perm_val]
        perm_shape = tuple(self.shape[p] for p in self.caching_perm_val)
        perm_back_d = {p: i for i, p in enumerate(self.caching_perm_val)}
        perm_back = [perm_back_d[i] for i in range(len(view))]

        out = self.get_canonical_cache_suffix_view(perm_view, perm_shape, perm_back, view)
        return op.map(out, lambda o: [o[p] for p in perm_back])

    def getView(self, view: ViewSpec, truncate_nan_rows: bool = False, allow_cache: bool = True):
        """
        view is list, where every element is either:
            - 'axis' or 'facet', representing no change
            - int, representing index for that dim
            - str representing reduction fn
        """
        print(f"{view=}")
        print(f"{os.getpid()=}")

        if self.caching and allow_cache and (canon_view := self.get_canonical_cache_suffix_view_perm(view)) is not None:
            cachable_canon_view = tuple((tuple(x) if isinstance(x, list) else x) for x in canon_view)
            if cachable_canon_view in self.cache:
                print("hit cache!!!", f"{canon_view=}")
                canon_vnt = self.cache[cachable_canon_view]
            else:
                print("fresh computation", f"{canon_view=}")
                canon_vnt = self.getView(canon_view, truncate_nan_rows=False, allow_cache=False)
                self.cache[cachable_canon_view] = canon_vnt

            ret_vnt = copy(canon_vnt)
            new_tensor = ret_vnt.tensor.clone()

            for i, (c_v_el, v_el) in enumerate(zip(canon_view, view)):
                if v_el == "mean" and c_v_el == "sum":
                    new_tensor /= self.circuit.shape[i]

            ret_vnt.tensor = new_tensor

            reduced_away = lambda x: x in ["mean", "sum"] or isinstance(x, int)
            # if the dim wasn't reduced and wasn't canonicalized, it must have been handled which implies that we should forward "axis"
            final_view: ViewSpec = [
                (v_el if c_v_el == "axis" else "axis")
                for c_v_el, v_el in zip(canon_view, view)
                if not reduced_away(c_v_el)
            ]

            return ret_vnt.getView(final_view, truncate_nan_rows=truncate_nan_rows)

        out_circ = self.circuit

        handled_dims = set[int]()
        removed_dims = set[int]()

        assert len(view) == self.ndim

        for view_el in view:
            if isinstance(view_el, str) and view_el not in ["mean", "sum", "axis"]:
                warn(
                    f"{view_el=}\nFor the circuit VNT, non mean/sum reduction operations are run last!\nThese "
                    "ops aren't commutative over ordering, so this isn't consistent with other VNT types!"
                )

        for i, view_el in reversed(list(enumerate(view))[: self.non_avoid_count]):
            maybe_new_circ_removed = self.apply_view_to_circ(out_circ, view_el, i)
            if maybe_new_circ_removed is not None:
                handled_dims.add(i)
                out_circ, is_removed = maybe_new_circ_removed
                if is_removed:
                    removed_dims.add(i)

        eval_out = self.eval_callback(out_circ, view[self.non_avoid_count :])

        handled: Set[int]
        removed: Set[int]
        if isinstance(eval_out, torch.Tensor):
            tensor = eval_out
            handled, removed = set(), set()
        else:
            tensor, handled, removed = eval_out
        assert handled.issubset(range(len(view[self.non_avoid_count :])))
        assert removed.issubset(range(len(view[self.non_avoid_count :])))
        for was_handled in handled:
            handled_dims.add(was_handled + self.non_avoid_count)
        for was_removed in removed:
            removed_dims.add(was_removed + self.non_avoid_count)

        assert handled_dims.issuperset(removed_dims)

        new_dim_idxs: List[List[str]] = []
        for dim_idxs, view_el in zip(self.dim_idx_names, view):
            if (idx := self.try_idx_view(view_el)) is not None and not isinstance(idx, int):
                dim_idxs = self.apply_idx_to_list(idx, dim_idxs)
            new_dim_idxs.append(list(dim_idxs))

        TypeVar("T")

        def filter_by_removed(item: Sequence[T]) -> List[T]:
            assert len(item) == self.ndim
            return [x for i, x in enumerate(item) if i not in removed_dims]

        return VeryNamedTensor(
            tensor,
            filter_by_removed(self.dim_names),
            filter_by_removed(self.dim_types),
            filter_by_removed(new_dim_idxs),
            units=self.units,
            title=self.title,
            aux_info=self.aux_info,
            selected_options=self.selected_options,
            default_view_spec=self.default_view_spec,
        ).getView(
            [x if i not in handled_dims else "axis" for i, x in enumerate(view) if i not in removed_dims],
            truncate_nan_rows=truncate_nan_rows,
        )

    def to_dict(self):
        print("sending circ vnt", self.title)
        return {
            "_getView": self.getView,
            "_getSparseView": lambda x: x,
            "dim_names": self.dim_names,
            "dim_idx_names": self.dim_idx_names,
            "dim_types": self.dim_types,
            "units": self.units,
            "title": self.title,
            "optionsSpec": None,
            "aux_info": self.aux_info,
            "default_view_spec": self.default_view_spec,
        }



import pytest

import rust_circuit as rc


def erroring_f(c):
    assert False


def thunk():
    pass


def c2c(c):
    return c


def ci2c(i, c):
    return c.rename(str(i))


def printy(c):
    print("found", c)


def test_python_callables():
    circuit = rc.Einsum.from_einsum_string("ab,bc->ac", rc.Array.randn(2, 3), rc.Array.randn(3, 4))
    print(circuit.get_compatible_device_dtype())
    print(circuit.map_children(c2c))
    print(circuit.map_children_enumerate(ci2c))

    with pytest.raises(AssertionError):
        print(circuit.map_children(erroring_f))
    with pytest.raises(TypeError):
        print(circuit.map_children(thunk))  # type: ignore
    circuit.visit(printy)
    with pytest.raises(TypeError):
        circuit.visit(c2c)  # type: ignore


if __name__ == "__main__":
    test_python_callables()



import functools
import json
import os
from dataclasses import dataclass
from typing import Any, Callable, List, Optional, Tuple, Union, cast

import torch

import rust_circuit.optional as op

RRFS_DIR = "/home/ubuntu/rrfs"
# from rust_circuit.rrfs import RRFS_DIR

# tokenizer loading copied from the old interpretability_tools.py to avoid jax dependency
RRFS_INTERP_MODELS_DIR = f"{RRFS_DIR}/interpretability_models_jax/"
MODELS_DIR = os.environ.get("INTERPRETABILITY_MODELS_DIR", RRFS_INTERP_MODELS_DIR)


def get_gpt_tokenizer_with_end_tok():
    import transformers

    tokenizer = transformers.GPT2TokenizerFast.from_pretrained(f"{MODELS_DIR}/gpt2/tokenizer")
    tokenizer._add_tokens(["[END]"])
    tokenizer.pad_token = "[END]"
    return tokenizer


@functools.cache
def get_interp_tokenizer():
    tokenizer = get_gpt_tokenizer_with_end_tok()
    tokenizer._add_tokens(["[BEGIN]"])
    tokenizer.eos_token = "[END]"
    return tokenizer


def with_dim_i_in_front(tensor, func: Callable[[Any], Any], axis):
    dim_permutation = [axis] + [i for i, x in enumerate(tensor.shape) if i != axis]
    inv_permutation = [x for x in dim_permutation]
    for i, x in enumerate(dim_permutation):
        inv_permutation[x] = i

    tensor = torch.permute(tensor, dim_permutation)
    tensor = func(tensor)
    tensor = torch.permute(tensor, inv_permutation)
    return tensor


def index_to_string(index):
    fixed_list = []
    for x in index:
        if isinstance(x, slice):
            if x.start is None and x.stop is None:
                fixed_list.append(":")
            else:
                fixed_list.append(
                    ("" if x.start is None else str(x.start)) + ":" + ("" if x.stop is None else str(x.stop))
                )
        elif isinstance(x, str):
            fixed_list.append(f'"{x}"')
        else:
            fixed_list.append(str(x))
    return f"[{', '.join(fixed_list)}]"


ViewSpecIdx = Union[int, str]
ViewSpec = List[Union[List[ViewSpecIdx], ViewSpecIdx]]


def array_getitem_separate(array, axes):
    """normal np getitem treats multiple array dimensions as matched, but this doesn't"""
    axes = axes + (slice(None, None, None),) * (len(array.shape) - len(axes))
    for i, ax in reversed(list(enumerate(axes))):
        array = array.__getitem__(tuple((slice(None, None),) * i + (ax,)))
    return array


def entropy_fn(tensor, dim):
    if torch.sum(tensor < 0) > 0:
        return torch.full_like(torch.sum(tensor, dim=dim), torch.nan)
    tensor = tensor / torch.nansum(tensor, dim=dim, keepdim=True)
    return -torch.nansum(tensor * torch.log(tensor))


def torch_take(tensor: torch.Tensor, indices, axis: int):
    return tensor[(slice(None, None),) * axis + (indices,)]


@dataclass
class VeryNamedTensor:
    reduction_fns = {
        "mean": torch.nanmean,
        "sum": torch.nansum,
        "norm": torch.linalg.norm,
        "max": torch.amax,
        "min": torch.amin,
        "entropy": entropy_fn,
    }

    def __init__(
        self,
        tensor,
        dim_names,
        dim_types,
        dim_idx_names,
        units="units",
        title="untitled_tensor",
        aux_info: Any = {},
        selected_options=None,
        default_axis_view: Union[List[ViewSpecIdx], ViewSpecIdx] = "axis",
        default_view_spec: Optional[ViewSpec] = None,  # overrides axis
    ):
        self.tensor: torch.Tensor = tensor
        self.selected_options = selected_options
        self.dim_names: List[str] = dim_names
        self.dim_types: List[str] = dim_types
        self.dim_idx_names: List[List[str]] = [[str(i) for i in x] for x in dim_idx_names]
        self.units: str = units
        self.title = title
        self.shape = self.tensor.shape
        self.aux_info = aux_info
        via_axis: ViewSpec = [default_axis_view] * len(self.dim_names)
        self.default_view_spec: ViewSpec = op.unwrap_or(default_view_spec, via_axis)

        assert len(dim_names) == len(
            self.shape
        ), f"{len(dim_names)} dim_names given ({dim_names}), but tensor is of rank {len(self.shape)} ({self.shape})"
        assert len(dim_types) == len(
            self.shape
        ), f"{len(dim_types)} dim_types given ({dim_types}), but tensor is of rank {len(self.shape)} ({self.shape})"
        idx_shape = tuple([len(g) for g in self.dim_idx_names])
        assert (
            idx_shape == self.shape
        ), f"dim_idx_names of shape {idx_shape} does not match tensor of shape {self.shape}"

    # can index with slices of numbers, or names
    # doesn't support
    def __getitem__(self, axes):
        return self.getitem(axes, False)

    def getitem(self, axes, no_title_change=False):
        if not isinstance(axes, tuple):
            axes = (axes,)

        if len(axes) < len(self.dim_types):
            axes = axes + (slice(None, None),) * (len(self.dim_types) - len(axes))

        axes = tuple([self.dim_idx_names[i].index(x) if isinstance(x, str) else x for i, x in enumerate(axes)])
        used_axes = [i for i, x in enumerate(axes) if (isinstance(x, slice) or isinstance(x, torch.Tensor))]
        dim_idx_names = []
        for i, (ax, names) in enumerate(zip(axes, self.dim_idx_names)):
            if i in used_axes:
                if isinstance(ax, torch.Tensor):
                    dim_idx_names.append([names[j] for j in ax])
                else:
                    dim_idx_names.append(names.__getitem__(ax))
        return VeryNamedTensor(
            tensor=array_getitem_separate(self.tensor, axes),
            dim_names=[name for i, name in enumerate(self.dim_names) if i in used_axes],
            dim_types=[ty for i, ty in enumerate(self.dim_types) if i in used_axes],
            dim_idx_names=dim_idx_names,
            units=self.units,
            title=self.title if no_title_change else f"{self.title}{index_to_string(axes)}",
            selected_options=self.selected_options,
            default_view_spec=self.default_view_spec,
        )

    def getView(self, view: ViewSpec, truncate_nan_rows=False):
        """
        view is list, where every element is either:
            - 'axis' or 'facet', representing no change
            - int, representing index for that dim
            - str in self.reduction_fns.keys(), representing reduction fn
        """
        result_tensor = self.tensor
        used_axes = [i for i, x in enumerate(view) if x in ["axis", "facet"] or isinstance(x, list)]
        for i, view_el in reversed(list(enumerate(view))):
            if view_el in ["axis", "facet"]:
                continue
            elif isinstance(view_el, str):
                assert (
                    view_el in self.reduction_fns.keys()
                ), f"element {view_el} in view {view} not supported, options are {list(self.reduction_fns.keys())}"
                reduction_fn: Callable = op.unwrap(self.reduction_fns.get(view_el))  # type: ignore
                result_tensor = reduction_fn(result_tensor, dim=i)
            elif isinstance(view_el, int):
                result_tensor = torch_take(result_tensor, view_el, i)
            elif isinstance(view_el, list):
                string_idxs = [i for i, x in enumerate(view_el) if isinstance(x, str)]
                idxs_zeros_for_strings = [-1 if isinstance(x, str) else x for x in view_el]
                view_el_v = view_el

                def myfunc(tensor):
                    tensor = tensor.clone()[torch.tensor(idxs_zeros_for_strings, dtype=torch.long)]
                    for string_idx in string_idxs:
                        view_item = view_el_v[string_idx]
                        assert isinstance(view_item, str)
                        tensor[string_idx] = self.reduction_fns[view_item](tensor, dim=0)
                    return tensor

                result_tensor = with_dim_i_in_front(result_tensor, myfunc, i)
                assert isinstance(result_tensor, torch.Tensor)
            else:
                raise ValueError(f"Invalid view: {view}")
        result_vnt = self.__getitem__(
            tuple(
                [
                    slice(None, None)
                    if dim in ["axis", "facet"]
                    else (
                        torch.tensor(
                            [inner_idx if isinstance(inner_idx, int) else 0 for inner_idx in dim], dtype=torch.long
                        )
                        if isinstance(dim, list)
                        else 0
                    )
                    for dim in view
                ]
            )
        )
        for used_i, base_i in enumerate(used_axes):
            listy_view_el = view[base_i]
            if isinstance(listy_view_el, list):
                result_vnt.dim_idx_names[used_i] = [
                    self.dim_idx_names[base_i][inner_idx] if isinstance(inner_idx, int) else inner_idx
                    for inner_idx in listy_view_el
                ]

        assert isinstance(result_tensor, torch.Tensor), type(result_tensor)

        result_vnt.tensor = result_tensor
        return result_vnt.truncate_nan_rows() if truncate_nan_rows else result_vnt

    def truncate_nan_rows(self):
        is_nan = torch.isnan(self.tensor)
        all_nan_by_dim_row = [
            torch.stack([torch_take(is_nan, torch.tensor([i]), dim).all() for i in range(self.shape[dim])], dim=0)
            for dim in range(len(self.shape))
        ]
        any_row_all_nan = torch.stack(all_nan_by_dim_row).any()
        if any_row_all_nan:
            new_tensor = self.tensor
            new_dim_idx_names = self.dim_idx_names
            for dim, rows_all_nan in enumerate(all_nan_by_dim_row):
                if rows_all_nan.any():
                    not_nan_idxs = torch.arange(len(rows_all_nan))[~rows_all_nan]
                    new_tensor = torch_take(new_tensor, not_nan_idxs, dim)
                    new_dim_idx_names[dim] = [self.dim_idx_names[dim][i] for i in not_nan_idxs]
            return VeryNamedTensor(
                tensor=new_tensor,
                dim_names=self.dim_names,
                dim_types=self.dim_types,
                dim_idx_names=new_dim_idx_names,
                units=self.units,
                title=self.title,
                selected_options=self.selected_options,
                default_view_spec=self.default_view_spec,
            )
        else:
            return self

    def to_dict(self):
        return {
            "tensor": self.tensor.cpu().numpy(),
            "dim_names": self.dim_names,
            "dim_idx_names": self.dim_idx_names,
            "dim_types": self.dim_types,
            "units": self.units,
            "title": self.title,
            "selected_options": self.selected_options,
            "default_view_spec": self.default_view_spec,
        }

    def to_lvnt(self):
        return LazyVeryNamedTensor(
            lambda: self.tensor,
            dim_names=self.dim_names,
            dim_types=self.dim_types,
            dim_idx_names=self.dim_idx_names,
            units=self.units,
            title=self.title,
            default_view_spec=self.default_view_spec,
        )

    def __repr__(self):
        indent = " " * 4  # len("VeryNamedTensor(")
        str = "VeryNamedTensor(\n"
        str += f"{indent}shape={self.shape}\n"
        str += f"{indent}dim_names={self.dim_names}\n"
        str += f"{indent}dim_types={self.dim_types}\n"
        str += f"{indent}dim_idx_names={self.dim_idx_names}\n"
        str += f"{indent}units={self.units}\n"
        str += f"{indent}tensor={self.tensor}\n"
        str += f"{indent}default_view_spec={self.default_view_spec}\n"
        str += ")"
        return str


class LazyVeryNamedTensor:
    def __init__(
        self,
        tensor_thunk,
        dim_names,
        dim_types,
        dim_idx_names,
        units="units",
        title="untitled_tensor",
        aux_info: Any = {},
        options_spec: Any = None,
        default_axis_view: Union[List[ViewSpecIdx], ViewSpecIdx] = "axis",
        default_view_spec: Optional[ViewSpec] = None,  # overrides axis
    ):
        self.options_spec = options_spec
        self.tensor_thunk: Union[Callable[[], Any], Callable[[Any], Any]] = tensor_thunk
        self.dim_names: List[str] = dim_names
        self.dim_types: List[str] = dim_types
        self.dim_idx_names: List[List[str]] = [[str(i) for i in x] for x in dim_idx_names]
        self.units: str = units
        self.title = title
        self.shape = tuple([len(x) for x in dim_idx_names])
        self.aux_info = aux_info
        self.realized_vnt: Optional[VeryNamedTensor] = None
        via_axis: ViewSpec = [default_axis_view] * len(self.dim_names)
        self.default_view_spec: ViewSpec = op.unwrap_or(default_view_spec, via_axis)

        assert len(dim_names) == len(
            self.shape
        ), f"{len(dim_names)} dim_names given ({dim_names}), but tensor is of rank {len(self.shape)} ({self.shape})"
        assert len(dim_types) == len(
            self.shape
        ), f"{len(dim_types)} dim_types given ({dim_types}), but tensor is of rank {len(self.shape)} ({self.shape})"
        idx_shape = tuple([len(g) for g in self.dim_idx_names])
        assert (
            idx_shape == self.shape
        ), f"dim_idx_names of shape {idx_shape} does not match tensor of shape {self.shape}"

    def __getitem__(self, axes):
        return self.getitem(axes, False)

    def getitem(self, axes: Tuple, no_title_change: bool = False):
        if not isinstance(axes, tuple):
            axes = (axes,) + (slice(None, None),) * (len(self.dim_types) - 1)

        if len(axes) < len(self.dim_types):
            axes = axes + (slice(None, None),) * (len(self.dim_types) - len(axes))

        axes = tuple([self.dim_idx_names[i].index(x) if isinstance(x, str) else x for i, x in enumerate(axes)])

        if self.realized_vnt is not None:
            thunk = cast(
                Union[Callable[[], Any], Callable[[Any], Any]],
                lambda: op.unwrap(self.realized_vnt).tensor.__getitem__(axes),
            )
        else:
            if self.options_spec is not None:
                # using type ignores bc mypy doesn't know what "function that takes 1 or 0 arguments" is
                thunk = lambda options: cast(Any, self.tensor_thunk)(fd.freeze(options)).__getitem__(axes)  # type: ignore
            else:
                thunk = lambda: cast(Any, self.tensor_thunk)().__getitem__(axes)  # type: ignore
        return LazyVeryNamedTensor(
            tensor_thunk=thunk,
            dim_names=[x for i, x in zip(axes, self.dim_names) if isinstance(i, slice)],
            dim_types=[x for i, x in zip(axes, self.dim_types) if isinstance(i, slice)],
            dim_idx_names=[x.__getitem__(i) for i, x in zip(axes, self.dim_idx_names) if isinstance(i, slice)],
            units=self.units,
            title=self.title if no_title_change else f"{self.title}{index_to_string(axes)}",
            aux_info=self.aux_info,
            options_spec=self.options_spec,
            default_view_spec=self.default_view_spec,
        )

    def getView(
        self,
        view,
        options=None,
        truncate_nan_rows=False,
    ):
        realized_tensor = None
        if self.realized_vnt is None or (
            self.options_spec is not None and json.dumps(options) != json.dumps(self.realized_vnt.selected_options)
        ):
            realized_tensor = self.tensor_thunk(*([options] if options is not None else []))
        if realized_tensor is not None:
            self.realized_vnt = VeryNamedTensor(
                tensor=realized_tensor,
                dim_names=self.dim_names,
                dim_types=self.dim_types,
                dim_idx_names=self.dim_idx_names,
                units=self.units,
                title=self.title + (" options:" + json.dumps(options) if options is not None else ""),
                selected_options=options,
            )
        return op.unwrap(self.realized_vnt).getView(view, truncate_nan_rows=truncate_nan_rows)

    def to_dict(self):
        return {
            "_getView": self.getView,
            "_getSparseView": lambda x: x,  # TODO reimplement this
            "dim_names": self.dim_names,
            "dim_idx_names": self.dim_idx_names,
            "dim_types": self.dim_types,
            "units": self.units,
            "title": self.title,
            "optionsSpec": self.options_spec,
            "aux_info": self.aux_info,
            "default_view_spec": self.default_view_spec,
        }


def vnt_guessing_shit_model_tokens(tensor, model, tokens, title="title", permissive=True, potential_dims=[]):
    return vnt_guessing_shit(
        tensor,
        [
            {"type": "seq", "name": "seq", "idx_names": [get_interp_tokenizer().decode([x]) for x in tokens]},
            {"type": "layer", "name": "layer", "idx_names": [str(i) for i in range(model.num_layers)]},
            {
                "type": "layerWithIO",
                "name": "layerWithIO",
                "idx_names": (["embeds"] + [str(i) for i in range(model.num_layers)] + ["outputs"]),
            },
            {"type": "heads", "name": "heads", "idx_names": [str(i) for i in range(model.num_heads)]},
            {"type": "neurons", "name": "neurons", "idx_names": [str(i) for i in range(model.hidden_size * 4)]},
            {"type": "hidden", "name": "hidden", "idx_names": [str(i) for i in range(model.hidden_size)]},
            *potential_dims,
        ],
        title=title,
        permissive=permissive,
    )


def vnt_guessing_shit(tensor, potential_dims, title="title", units="units", permissive=True):
    dim_types = []
    dim_idx_names = []
    dim_names = []
    for dim_length in tensor.shape:
        matching_potentials = [x for x in potential_dims if len(x["idx_names"]) == dim_length]
        if len(matching_potentials) == 0:
            assert permissive is True, "no matching length dims"
            dim_idx_names.append([str(i) for i in range(dim_length)])
            dim_types.append("unknown")
            name = "unknown"
            while name in dim_names:
                name += "again"
            dim_names.append(name)
        else:
            matching = matching_potentials[0]
            dim_types.append(matching["type"])
            dim_idx_names.append(matching["idx_names"])
            name = matching["name"]
            while name in dim_names:
                name += "again"
            dim_names.append(name)
    return VeryNamedTensor(tensor, dim_names, dim_types, dim_idx_names, units, title)


def get_shapes_spec(tokens, model):
    return {
        "tokens": tokens,
        "numLayers": model.num_layers,
        "numHeads": model.num_heads,
        "hiddenSize": model.hidden_size,
    }



import rust_circuit as rc

P = rc.Parser()


def test_circuit_flags():
    s = """
    'a' [3] Symbol
    'b' Einsum ij,j->j
      'c' [7, 3] Scalar 2.
      'a'
    'b_arr' Einsum ij,j->j
      'c_arr' [7, 3] Array rand
      'a_arr' [3] Array rand
    'b_var' DiscreteVar
      'vals' [7, 3] Array rand
      'probs' GeneralFunction softmax
        'x' [7] Array rand
    'b_var_sym' DiscreteVar
      'vals_sym' [7, 3] Symbol
      'probs'
    'cum_var' Cumulant
      'b_var'
    'cum_var_sym' Cumulant
      'b_var_sym'
    'cum_sym' Cumulant
      'a'
    'cum_empty' Cumulant
    'cum_const' Cumulant
      'b_arr'
      'c_arr'
    """
    (
        a,
        b,
        b_arr,
        b_var,
        b_var_sym,
        cum_var,
        cum_var_sym,
        cum_sym,
        cum_empty,
        cum_const,
    ) = P.parse_circuits(s)

    def check(
        c: rc.Circuit, is_constant: bool, is_explicitly_computable: bool, can_be_sampled: bool, use_autoname: bool
    ):
        assert c.is_constant == is_constant
        assert c.is_explicitly_computable == is_explicitly_computable
        assert c.can_be_sampled == can_be_sampled
        assert c.use_autoname == use_autoname

        sym = rc.Symbol.new_with_random_uuid((), name="sym")
        mod = rc.module_new_bind(sym, (sym, c))
        assert mod.is_constant == c.is_constant
        assert mod.is_explicitly_computable == c.is_explicitly_computable
        assert mod.can_be_sampled == c.can_be_sampled
        assert mod.no_unbound_symbols == c.no_unbound_symbols
        assert mod.use_autoname == c.use_autoname

    check(a, is_constant=True, is_explicitly_computable=False, can_be_sampled=True, use_autoname=True)
    check(b, is_constant=True, is_explicitly_computable=False, can_be_sampled=True, use_autoname=True)
    check(b_arr, is_constant=True, is_explicitly_computable=True, can_be_sampled=True, use_autoname=True)
    check(b_var, is_constant=False, is_explicitly_computable=False, can_be_sampled=True, use_autoname=True)
    check(b_var_sym, is_constant=False, is_explicitly_computable=False, can_be_sampled=True, use_autoname=True)
    check(cum_var, is_constant=True, is_explicitly_computable=False, can_be_sampled=True, use_autoname=True)
    check(cum_var_sym, is_constant=True, is_explicitly_computable=False, can_be_sampled=True, use_autoname=True)
    check(cum_sym, is_constant=True, is_explicitly_computable=False, can_be_sampled=True, use_autoname=True)
    check(cum_empty, is_constant=True, is_explicitly_computable=True, can_be_sampled=True, use_autoname=True)
    check(cum_const, is_constant=True, is_explicitly_computable=True, can_be_sampled=True, use_autoname=True)

    a_dis = a.with_autoname_disabled()
    check(a_dis, is_constant=True, is_explicitly_computable=False, can_be_sampled=True, use_autoname=False)
    b_arr_dis = b_arr.update("c_arr", lambda x: x.with_autoname_disabled())
    check(b_arr_dis, is_constant=True, is_explicitly_computable=True, can_be_sampled=True, use_autoname=False)

    assert rc.Add(b_arr_dis).op_name is None
    assert rc.Add(b_arr).op_name is not None



import pytest

import rust_circuit as rc


# TODO: more tests
def test_symbolic_expand_simple():
    P = rc.Parser()

    scl, v, new = P.parse_circuits(
        """
    1 [7] Scalar 1.3
    0 Einsum i,i->i
      1
      2 [7] Scalar 1.5
    'new' [?] Symbol
    """
    )

    assert rc.expand_node(v, [scl, new]).shape == (7,)


def test_symbolic_rearrange_expand():
    P = rc.Parser()
    scl, sym, rearrange = P.parse_circuits(
        """
    0 [6] Scalar 1.3
    'sym' [?] Symbol rand
    1 Rearrange (a:2 b:3) -> b a
      0
    """
    )

    # should use set symbolic
    expanded = rc.Expander((scl, lambda _: sym))(rearrange)
    assert expanded.shape == (3, 2)

    scl0, scl1, scl2, sym, rearrange = P.parse_circuits(
        """
    'scl0' [?] Symbol
    'scl1' [24] Scalar 1.3
    'scl2' [17] Scalar 1.3
    'sym' [?] Symbol rand
    2 Add
      3 Rearrange (a b:3) -> b a
        'scl0'
      4 Rearrange (b:3 a:8) -> a b 1
        'scl1'
      8 Einsum i->
        7 Add
          5 Rearrange a:17 -> a
            'scl2'
          6 Rearrange a -> a
            'scl2'
    """
    )

    expanded = rc.Expander(({scl0, scl1, scl2}, lambda _: sym))(rearrange)
    s0, *_ = rc.symbolic_sizes()
    assert expanded.shape == (8, 3, s0)


def test_multiple_symbolic_product():
    s = """
    'a' [?] Symbol 07a8025a-d8ba-4c57-946d-cd7cc7e78a66
    'b' Rearrange a b c -> (a b c)
      'c' [?, ?, ?] Symbol 0b108f80-b37f-4e2c-821f-45cbedcbde75
    'd' Rearrange a c -> (a c)
      'e' [?, ?] Symbol 3f8aaf28-58f8-44fc-8dbc-497844a5a837
    """
    a, b, d = rc.Parser().parse_circuits(s)
    assert a.repr() == "'a' [?] Symbol 07a8025a-d8ba-4c57-946d-cd7cc7e78a66"
    assert (
        rc.PrintOptions(shape_only_when_necessary=False).repr(b)
        == "'b' [?] Rearrange a b c -> (a b c)\n  'c' [?,?,?] Symbol 0b108f80-b37f-4e2c-821f-45cbedcbde75"
    )
    assert (
        rc.PrintOptions(shape_only_when_necessary=False).repr(d)
        == "'d' [?] Rearrange a b -> (a b)\n  'e' [?,?] Symbol 3f8aaf28-58f8-44fc-8dbc-497844a5a837"
    )



from __future__ import annotations

from typing import Dict, Iterable, Mapping, Optional, Set, Type, TypeVar, Union

import attrs
import torch
from attr import frozen

import rust_circuit as rc

from ..py_utils import FrozenDict


def all_same_len(self: Dataset, attribute, arrs: Mapping[str, rc.Array]):
    assert len(arrs) >= 1
    l = None
    for a in arrs.values():
        if l is None:
            l = a.shape[0]
        elif not a.shape[0] == l:
            raise ValueError(
                f"All Arrays in a Dataset must have the same `len`, found instead at least two: {l}, {a.shape[0]}"
            )


def arrs_to_map(arrs: Union[Iterable[rc.Array], Mapping[str, rc.Array]]) -> FrozenDict[str, rc.Array]:
    if isinstance(arrs, Mapping):
        for name, a in arrs.items():
            assert name == a.name
        return FrozenDict(arrs)
    else:
        arrs_map: Dict[str, rc.Array] = {}
        for a in arrs:
            if a.name in arrs_map:
                if not a == arrs_map[a.name]:
                    raise ValueError(f"We found two different Arrays with the same name: {a}, {arrs_map[a.name]}")
            arrs_map[a.name] = a
        return FrozenDict(arrs_map)


def names_in_arrs(self: Dataset, attribute, input_names: Set[str]):
    for name in input_names:
        assert name in self.arrs, name


def frozenset_converter(xs: Set[str]) -> frozenset[str]:  # mypy has issues with just using frozenset as converter
    return frozenset(xs)


TDataset = TypeVar("TDataset", bound="Dataset")


@frozen(hash=True)
class Dataset:
    # Holds arrays look-up-able by name; you can provide any iterable and it will be turned into a map.
    # In particular no two can have the same name. Also, they can be accessed as `ds.arr_name`, unless
    # you name your array something silly like __len__
    arrs: Mapping[str, rc.Array] = attrs.field(validator=all_same_len, converter=arrs_to_map)
    # Which arrays (by name) are inputs to the circuit that should be replaced when doing causal scrubbing.
    # By default this is all of them.
    input_names: Set[str] = attrs.field(
        validator=names_in_arrs, factory=set, converter=frozenset_converter, kw_only=True
    )

    def __attrs_post_init__(self):
        if not len(self.input_names):
            object.__setattr__(self, "input_names", frozenset(self.arrs.keys()))

    def __getattr__(self, __name: str) -> rc.Array:
        try:
            return self.arrs[__name]
        except KeyError:
            raise AttributeError(__name)

    def __len__(self) -> int:
        return list(self.arrs.values())[0].shape[0]

    def __getitem__(self, idxs: rc.TorchAxisIndex):
        if isinstance(idxs, int):
            idxs = slice(idxs, idxs + 1)
        return attrs.evolve(self, arrs={name: rc.Array(inp.value[idxs], name) for name, inp in self.arrs.items()})

    def __str__(self) -> str:
        # Probably you want to overwrite this when subclassing so you get pretty prints!
        return str([f"<{a.name} {a.shape}>" for a in self.arrs.values()])

    def sample(self, count, rng: Optional[torch.Generator] = None):
        idxs = torch.multinomial(
            torch.ones(size=(len(self),), dtype=torch.float32, device=(rng.device if rng else None)),
            num_samples=count,
            replacement=True,
            generator=rng,
        )
        return self[idxs]

    @classmethod
    def unwrap(cls: Type[TDataset], d: Dataset) -> TDataset:
        assert isinstance(d, cls), (type(d), cls)
        return d


def color_dataset(ds, html=False):
    def color_string_from_int(i: int):
        hue = (i % (255 / 10)) * 10
        return f"hsl({hue}, 90%, 60%)"

    if html:
        return "darkgrey" if ds is None else color_string_from_int(hash(ds))
    else:
        # pretty made up, would be nice if printer had a "color by feature"
        return 6 if ds is None else 2 + hash(ds) % 6



import itertools
from typing import List, Type, Union

import pytest

import rust_circuit as rc
import rust_circuit.optional as op
from interp.tools.indexer import SLICER as S
from rust_circuit import (
    FINISHED,
    Add,
    AnyFound,
    Array,
    Circuit,
    Cumulant,
    Einsum,
    Expander,
    Finished,
    GeneralFunction,
    Getter,
    IterateMatchResults,
    IterativeMatcher,
    IterativeMatcherChildNumbersOutOfBoundsError,
    Matcher,
    Module,
    PrintOptions,
    Regex,
    Scalar,
    Symbol,
    Transform,
    Updater,
    all_children,
    append_matchers_to_names,
    module_new_bind,
    new_traversal,
    print_matcher_debug,
    restrict,
    restrict_sl,
)
from rust_circuit.module_library import TransformerBlockParams, layernorm


# By default I'd recommend using the multi argument constructor for doing
# 'any'. We don't do always do this here, but that's for testing reasons.
def test_basic_matching():
    a = Scalar(1.0, shape=(1, 2), name="a_name")
    b = Scalar(2.0, shape=(2, 3), name="b_other")
    c = Einsum.from_einsum_string("ab,bc->ac", a, b)

    assert Matcher("a_name")(a)
    assert not Matcher("a_name")(b)
    assert Matcher({"a_name", "b_other"})(a)
    assert Matcher(frozenset[str]({"a_name", "b_other"}))(a)
    assert Matcher(frozenset[Type[Circuit]]({Scalar, Einsum}))(c)
    assert Matcher(frozenset[Circuit]({c, b}))(c)
    assert Matcher({"a_name", "b_other"})(b)
    assert not Matcher({"a_name", "b_other"})(c)

    assert Matcher(a)(a)
    assert not Matcher(a)(b)
    assert Matcher(set[Circuit]({a, c}))(a)
    assert Matcher(set[Circuit]({a, c}))(c)
    assert not Matcher(set[Circuit]({a, c}))(b)
    assert Matcher(Scalar)(b)
    assert not Matcher(Scalar)(c)
    assert not Matcher(Scalar).call(c)
    assert Matcher(set[Type[Circuit]]({Scalar, Einsum}))(c)
    assert Matcher(set[Type[Circuit]]({Scalar, Einsum}))(a)
    assert Matcher(set[Type[Circuit]]({Scalar, Einsum})).call(a)

    assert Matcher.circuits(a)(a)
    assert not Matcher.circuits(a)(b)
    assert Matcher.circuits(a, c)(a)
    assert Matcher.circuits(a, c)(c)
    assert not Matcher.circuits(a, c)(b)

    has_child = Matcher.match_any_child_found(a)
    assert not has_child(a)
    assert not has_child(b)
    assert has_child(c)
    assert has_child == Matcher.match_any_child_found(a)
    assert has_child == Matcher.match_any_child_found(set[Circuit]({a}))
    has_child = Matcher.match_any_child_found(set[Circuit]({a, c}))
    assert not has_child(a)
    assert not has_child(b)
    assert has_child(c)
    assert has_child == Matcher.match_any_child_found(set[Circuit]({c, a}))
    e = Scalar(3.0, shape=(1, 3), name="hi")
    d = Add(c, e)
    has_child = Matcher.match_any_child_found(a)
    assert has_child(d)
    assert has_child(c)
    assert not has_child(a)
    assert not has_child(b)
    assert not has_child(e)

    match_any_found = Matcher.match_any_found(a)
    assert match_any_found(d)
    assert match_any_found(c)
    assert match_any_found(a)
    assert not match_any_found(b)
    assert not match_any_found(e)

    assert Matcher.types(Scalar)(b)
    assert not Matcher.types(Scalar)(c)
    assert Matcher.types(Scalar, Einsum)(c)
    assert Matcher.types(Scalar, Einsum)(a)
    assert not Matcher.types(Add, Einsum)(a)

    assert Matcher("a_name") == Matcher("a_name")
    assert Matcher("a_name") != Matcher("a_name_")
    assert Matcher(Scalar) == Matcher(Scalar)
    assert Matcher.types(Scalar, Einsum) == Matcher.types(Einsum, Scalar)
    assert Matcher.types(Scalar, Einsum) == Matcher(set[Type[Circuit]]({Einsum, Scalar}))
    assert Matcher(Scalar) != Matcher(Einsum)
    assert Matcher(Scalar) != Matcher(Einsum)
    assert Matcher.true_matcher() == Matcher(True)
    assert Matcher.false_matcher() == Matcher(False)
    assert Matcher.false_matcher() != Matcher(True)
    assert hash(Matcher.false_matcher()) == hash(Matcher(False))

    assert Matcher("a_name").new_or("b_other")(b)
    assert Matcher("a_name").new_or("b_other")(a)
    assert not Matcher("a_name").new_or("b_other")(c)
    assert not Matcher("a_name").new_and("b_other")(a)
    assert not (Matcher("a_name") & "b_other")(a)
    assert (Matcher("a_name") | "b_other")(a)
    assert not ("b_other" & Matcher("a_name"))(a)
    assert ("b_other" | Matcher("a_name"))(a)
    assert not (~Matcher("a_name"))(a)
    assert (~Matcher("a_name"))(b)
    assert (~Matcher("a_name"))(c)
    assert not (Matcher("a_name").new_not())(a)
    assert (Matcher("a_name").new_not())(b)
    assert (Matcher("a_name").new_not())(c)
    assert Matcher("a_name").new_and(Scalar)(a)
    assert not Matcher("a_name").new_and(Scalar)(b)

    assert Matcher.any("a_name", "b_other")(b)
    assert Matcher.any("a_name", "b_other")(a)
    assert not Matcher.any("a_name", "b_other")(c)
    assert Matcher("a_name", "b_other") == Matcher.any("a_name", "b_other")

    # matcher short-circuiting
    empty_arr = []
    assert not (Matcher("a_name") & Matcher(lambda _: empty_arr[0]))(b)
    assert not Matcher.all("a_name", lambda _: empty_arr[0])(b)
    assert Matcher.any("b_other", lambda _: empty_arr[0])(b)
    assert (Matcher("b_other") | Matcher(lambda _: empty_arr[0]))(b)

    assert Matcher.all("a_name", Scalar)(a)
    assert not Matcher.all("a_name", Scalar)(b)
    assert not Matcher.all("a_name", Scalar)(c)

    assert Matcher.all("a_name")(a)
    assert not Matcher.all("a_name")(b)
    assert not Matcher.all("a_name")(c)

    # it's preferred to use Matcher.regex in most cases
    assert Matcher.regex("a_name") == Matcher(Regex("a_name"))
    assert Matcher.regex("a_name", escape_dot=False) == Matcher(Regex("a_name", escape_dot=False))
    assert Matcher.regex("a_name") != Matcher.regex("b_name")
    assert Matcher.regex("a_name") == Matcher.regex("a_name")
    assert Matcher.regex("a_name") != Matcher.regex("a_name", escape_dot=False)
    assert Matcher.regex("a_name", escape_dot=False) == Matcher.regex("a_name", escape_dot=False)
    assert Matcher.regex("...a_name", escape_dot=False) == Matcher.regex("...a_name", escape_dot=False)
    assert Matcher.regex("...a_name") == Matcher.regex("...a_name")

    assert Matcher.regex("a")(a)
    assert not Matcher.regex("a")(b)

    assert Matcher.regex("[ab]")(a)
    assert Matcher.regex("[ab]")(b)

    assert Matcher.regex("^my.name.has_dot$")(Scalar(2.0, shape=(2, 3), name="my.name.has_dot"))
    assert not Matcher.regex("^my.name...s_dot$")(Scalar(2.0, shape=(2, 3), name="my.name.has_dot"))
    assert Matcher.regex(r"^my.name.\.\.s_dot$")(Scalar(2.0, shape=(2, 3), name="my.name.has_dot"))
    assert Matcher.regex(r"^\.*$")(c)
    assert not Matcher.regex(r"^\.*$")(c.rename(None))

    for n in [a, b, c]:
        for v in [False, True]:
            assert Matcher(v)(n) == v

    with pytest.raises(TypeError, match="coerced"):
        Matcher("a_name") and Matcher("b_other")


def test_matcher_get():
    a = Scalar(1.0, shape=(1, 2), name="a_name")
    b = Scalar(2.0, shape=(2, 3), name="b_other")
    c = Einsum.from_einsum_string("ab,bc->ac", a, b)

    assert Matcher("a_name").get(c) == {a}
    assert c.get("a_name") == {a}
    assert Matcher("b_other").get(c) == {b}
    assert Matcher("b_other").get_first(c) == b
    assert Matcher("a_name", "b_other").get_first(c) == b


def test_iterative_just_matcher():
    assert Matcher("hi").to_iterative_matcher() == IterativeMatcher("hi")
    assert Matcher("hi").to_iterative_matcher() != IterativeMatcher("hij")
    assert Matcher(False).to_iterative_matcher() == IterativeMatcher(False)
    assert Matcher(False).to_iterative_matcher() != IterativeMatcher(True)
    assert Matcher.regex("a").to_iterative_matcher() == IterativeMatcher(Matcher.regex("a"))
    assert Matcher.regex("a").to_iterative_matcher() != IterativeMatcher(Matcher.regex("ab"))
    assert Matcher(Matcher.regex("a")).to_iterative_matcher() != IterativeMatcher(Matcher.regex("ab"))

    a = Scalar(1.0, shape=(1, 2), name="a_name")
    b = Scalar(2.0, shape=(2, 3), name="b_other")
    c = Einsum.from_einsum_string("ab,bc->ac", a, b)

    def simple_match_check(res: IterateMatchResults, found: bool):
        assert res.updated is None
        assert res.found == found

    simple_match_check(IterativeMatcher("a_name").match_iterate(a), True)
    simple_match_check(IterativeMatcher("a_name").match_iterate(b), False)
    simple_match_check(IterativeMatcher({"a_name", "b_other"}).match_iterate(a), True)
    simple_match_check(IterativeMatcher({"a_name", "b_other"}).match_iterate(b), True)
    simple_match_check(IterativeMatcher({"a_name", "b_other"}).match_iterate(c), False)

    simple_match_check(IterativeMatcher(a).match_iterate(a), True)
    simple_match_check(IterativeMatcher(a).match_iterate(b), False)
    simple_match_check(IterativeMatcher(set[Circuit]({a, c})).match_iterate(a), True)
    simple_match_check(IterativeMatcher(set[Circuit]({a, c})).match_iterate(c), True)
    simple_match_check(IterativeMatcher(set[Circuit]({a, c})).match_iterate(b), False)
    simple_match_check(IterativeMatcher(Scalar).match_iterate(b), True)
    simple_match_check(IterativeMatcher(Scalar).match_iterate(c), False)
    simple_match_check(IterativeMatcher(Scalar).match_iterate(c), False)
    simple_match_check(IterativeMatcher(set[Type[Circuit]]({Scalar, Einsum})).match_iterate(c), True)
    simple_match_check(IterativeMatcher(set[Type[Circuit]]({Scalar, Einsum})).match_iterate(a), True)
    simple_match_check(IterativeMatcher(set[Type[Circuit]]({Scalar, Einsum})).match_iterate(a), True)


def test_iterative_restrict_constructor():
    all_items = list(
        itertools.product(
            ["hi", "x", True],
            [False, True],
            [None, 0, 2],
            [None, 0, 2],
            ["hi", False],
        )
    )

    for base, term_if_matches, start_depth, end_depth, term_early_at in all_items:
        restrict_args = term_if_matches, start_depth, end_depth, term_early_at
        restrict_sl_args = term_if_matches, slice(start_depth, end_depth), term_early_at
        item = restrict(base, *restrict_args)
        assert restrict(Matcher(base).to_iterative_matcher(), *restrict_args) == item
        assert restrict(IterativeMatcher(base), *restrict_args) == item
        assert restrict_sl(base, *restrict_sl_args) == item
        assert restrict_sl(IterativeMatcher(base), *restrict_sl_args) == item

        for (
            base_other,
            term_if_matches_other,
            start_depth_other,
            end_depth_other,
            term_early_at_other,
        ) in all_items:
            restrict_args_other = term_if_matches_other, start_depth_other, end_depth_other, term_early_at_other
            if base_other != base or restrict_args_other != restrict_args:
                other_item = restrict(Matcher(base_other), *restrict_args_other)
                assert other_item != item


def test_finished_eq():
    assert IterativeMatcher.term().match_iterate(Scalar(0.3)).updated == FINISHED


def test_iterative_restrict():
    m = restrict_sl("a_name", depth_slice=S[2:])
    a = Scalar(1.0, shape=(1, 2), name="a_name")
    b = Scalar(2.0, shape=(2, 3), name="b_other")

    def check_valid(m: IterativeMatcher, a: Circuit, reached: bool = True) -> IterativeMatcher:
        x = m.match_iterate(a)
        assert x.updated is None or isinstance(x.updated, IterativeMatcher)
        assert (x.updated is None) == reached, (x.updated, reached)
        assert x.found == reached

        return op.unwrap_or(x.updated, m)

    m = check_valid(m, a, reached=False)
    m = check_valid(m, a, reached=False)
    m = check_valid(m, a, reached=True)
    m = check_valid(m, a, reached=True)

    def check_valid_both(m: IterativeMatcher, a: Circuit, found: bool = True, last: bool = False):
        x = m.match_iterate(a)
        assert (x.updated == FINISHED) == last
        assert x.found == found
        if x.updated == FINISHED:
            # hack for typing
            return m

        assert x.updated is None or isinstance(x.updated, IterativeMatcher)
        return op.unwrap_or(x.updated, m)

    m = restrict_sl("a_name", depth_slice=S[3:5])

    for _ in range(3):
        m = check_valid_both(m, a, found=False)
    m = check_valid_both(m, a, found=True)
    m = check_valid_both(m, a, found=True, last=True)

    m = restrict_sl("a_name", depth_slice=S[:0])
    check_valid_both(m, a, found=False, last=True)
    m = restrict_sl("a_name", depth_slice=S[0:0])
    check_valid_both(m, a, found=False, last=True)
    m = restrict_sl("a_name", depth_slice=S[1:0])
    check_valid_both(m, a, found=False, last=True)

    m = restrict_sl("a_name", term_if_matches=True, depth_slice=S[3:5])
    for _ in range(3):
        m = check_valid_both(m, a, found=False)
    m = check_valid_both(m, a, found=True, last=True)

    m = restrict_sl("a_name", term_if_matches=True, depth_slice=S[3:7])
    for _ in range(3):
        m = check_valid_both(m, a, found=False)
    m = check_valid_both(m, b, found=False, last=False)
    m = check_valid_both(m, b, found=False, last=False)
    m = check_valid_both(m, a, found=True, last=True)

    def correct_no_update(m: IterativeMatcher, a: Circuit, found: bool = True, last: bool = False):
        x = m.match_iterate(a)
        if last:
            assert x.updated == FINISHED
            return m  # typing hack
        else:
            assert x.updated is None
        assert x.found == found
        assert x.updated is None or isinstance(x.updated, IterativeMatcher)

        return op.unwrap_or(x.updated, m)

    m = restrict("a_name", term_early_at="b_other")
    m = correct_no_update(m, a, found=True)
    m = correct_no_update(m, a, found=True)
    m = correct_no_update(m, b, found=False, last=True)


def test_start_depth_behavior():
    b = rc.Add(name="B")
    c = rc.Add(name="C")
    a = rc.Add(b, c, name="A")

    assert new_traversal(term_early_at=rc.Add, start_depth=0).get(a) == {a}
    for d in range(1, 4):
        assert new_traversal(term_early_at=rc.Add, start_depth=d).get(a) == set()

    assert new_traversal(start_depth=1, end_depth=0).match_iterate(a).updated == FINISHED


def test_chain_constructor():
    all_items = list(
        itertools.product(
            ["hi", "x", True],
            [("other",), ("other", "hi"), (False,), (False, False), (False, True)],
        )
    )

    for must_be_sub in [False, True]:
        for first, rest in all_items:
            for first_other, rest_other in all_items:
                item = Matcher(first).chain(*rest, must_be_sub=must_be_sub)
                assert Matcher(first).to_iterative_matcher().chain(*rest, must_be_sub=must_be_sub) == item
                assert IterativeMatcher.new_chain(first, *rest, must_be_sub=must_be_sub) == item

                if first_other != first or rest != rest_other:
                    other_item = Matcher(first_other).chain(*rest_other, must_be_sub=must_be_sub)
                    assert other_item != item

                assert Matcher(first).chain_many(rest, must_be_sub=must_be_sub) == item
                assert Matcher(first).to_iterative_matcher().chain_many(rest, must_be_sub=must_be_sub) == item
                assert IterativeMatcher.new_chain_many([first, *rest], must_be_sub=must_be_sub) == item


@pytest.mark.parametrize("must_be_sub", [True, False])
def test_chain_matching(must_be_sub: bool):
    # must_be_sub should make no difference for these tests
    m = IterativeMatcher.new_chain("top", "leaf", must_be_sub=must_be_sub)

    def check_update(m: IterativeMatcher, a: Circuit, found: bool = True, change: bool = True, finish: bool = False):
        x = m.match_iterate(a)
        if finish:
            assert x.updated == FINISHED
            return m  # hack for typing
        assert (x.updated is not None) == (change and not finish)
        assert x.found == found
        assert x.updated is None or isinstance(x.updated, IterativeMatcher)

        return op.unwrap_or(x.updated, m)

    a = Scalar(1.0, shape=(1, 2), name="top")
    b = Scalar(2.0, shape=(2, 3), name="leaf")
    c = Scalar(2.0, shape=(2, 3), name="other")
    d = Scalar(2.0, shape=(2, 3), name="dog")

    m = check_update(m, a, found=False)
    m = check_update(m, a, found=False, change=False)
    m = check_update(m, a, found=False, change=False)
    m = check_update(m, a, found=False, change=False)
    m = check_update(m, b, found=True, change=False)
    m = check_update(m, b, found=True, change=False)

    m = IterativeMatcher.new_chain(restrict("top", end_depth=2), restrict("leaf", end_depth=3), must_be_sub=must_be_sub)

    m = check_update(m, a, found=False)
    m = check_update(m, a, found=False)
    m = check_update(m, a, found=False)
    m = check_update(m, a, found=False, finish=True)

    m = IterativeMatcher.new_chain(restrict("top", end_depth=2), restrict("leaf", end_depth=3), must_be_sub=must_be_sub)

    m = check_update(m, b, found=False)
    m = check_update(m, b, found=False, finish=True)

    m = IterativeMatcher.new_chain(restrict("top", end_depth=2), restrict("leaf", end_depth=4), must_be_sub=must_be_sub)

    m = check_update(m, a, found=False)
    m = check_update(m, c, found=False)
    m = check_update(m, b, found=True)
    m = check_update(m, b, found=True, finish=True)

    m = IterativeMatcher.new_chain(
        restrict("top", end_depth=2),
        restrict("leaf", end_depth=4),
        Matcher("other"),
        must_be_sub=must_be_sub,
    )

    m = check_update(m, a, found=False)
    m = check_update(m, c, found=False)
    m = check_update(m, c, found=False)
    m = check_update(m, b, found=False)
    m = check_update(m, c, found=True, change=False)
    m = check_update(m, c, found=True, change=False)

    m = IterativeMatcher.new_chain_many(
        (restrict("top", end_depth=2), restrict("leaf", end_depth=4), Matcher("other")),
        (Matcher("dog"),),
        must_be_sub=must_be_sub,
    )

    m = check_update(m, a, found=False)
    m = check_update(m, d, found=True)
    m = check_update(m, c, found=False)
    m = check_update(m, b, found=False)
    m = check_update(m, c, found=True, change=False)
    m = check_update(m, c, found=True, change=False)
    m = check_update(m, d, found=True, change=False)

    m = IterativeMatcher.any(restrict("top", end_depth=2), Matcher("dog"))

    m = check_update(m, a, found=True)
    m = check_update(m, a, found=True)
    m = check_update(m, a, found=False, change=False)
    m = check_update(m, d, found=True, change=False)


def test_chain_must_be_sub():
    a = Scalar(1.0, shape=(1, 2), name="top")
    b = Scalar(2.0, shape=(2, 2), name="leaf")
    c = Scalar(2.0, shape=(2, 3), name="other")
    d = Scalar(2.0, shape=(2, 3), name="dog")

    def check_update(m: IterativeMatcher, a: Circuit, found: bool = True, change: bool = True, finish: bool = False):
        x = m.match_iterate(a)
        if finish:
            assert x.updated == FINISHED
            return m  # hack for typing
        assert (x.updated is not None) == (change and not finish)
        assert x.found == found
        assert x.updated is None or isinstance(x.updated, IterativeMatcher)

        return op.unwrap_or(x.updated, m)

    m = IterativeMatcher(Scalar).chain(Scalar, must_be_sub=False)
    m = check_update(m, a)
    m = check_update(m, b, change=False)
    m = check_update(m, c, change=False)

    m = IterativeMatcher(Scalar).chain(Scalar, must_be_sub=True)
    m = check_update(m, a, found=False)
    m = check_update(m, b, change=False)
    m = check_update(m, c, change=False)

    m = IterativeMatcher("top").chain(restrict(lambda x: x.shape[-1] == 2, end_depth=2), "dog", must_be_sub=True)
    m = check_update(m, b, found=False, change=False)
    m = check_update(m, d, found=False, change=False)
    m = check_update(m, a, found=False, change=True)
    m = check_update(m, a, found=False)
    m = check_update(m, d, found=True, change=True)

    m = IterativeMatcher("top").chain(restrict(lambda x: x.shape[-1] == 2, end_depth=2), "dog", must_be_sub=True)
    m = check_update(m, b, found=False, change=False)
    m = check_update(m, d, found=False, change=False)
    m = check_update(m, a, found=False, change=True)
    m = check_update(m, b, found=False)
    m = check_update(m, b, found=False, change=False)
    m = check_update(m, b, found=False, change=False)
    m = check_update(m, b, found=False, change=False)
    m = check_update(m, d, found=True, change=False)

    m = IterativeMatcher("top").chain(restrict(lambda x: x.shape[-1] == 2, end_depth=2), "dog", must_be_sub=True)
    m = check_update(m, b, found=False, change=False)
    m = check_update(m, d, found=False, change=False)
    m = check_update(m, a, found=False, change=True)
    m = check_update(m, c, found=False, change=True)
    m = check_update(m, c, found=False, change=False)
    m = check_update(m, d, found=False, change=False)

    m = IterativeMatcher("top").chain(restrict(lambda x: x.shape[-1] == 2, end_depth=2), "dog", must_be_sub=False)
    m = check_update(m, b, found=False, change=False)
    m = check_update(m, d, found=False, change=False)
    m = check_update(m, a, found=False, change=True)
    m = check_update(m, c, found=False, change=True)
    m = check_update(m, c, found=False, change=False)
    m = check_update(m, d, found=True, change=False)


def test_update_branching():
    a = Scalar(1.0, shape=(1, 2), name="a")
    b = Scalar(2.0, shape=(2, 3), name="b")
    ein = Einsum.from_einsum_string("ij,jk->ik", a, b, name="ein")
    soft = GeneralFunction.new_by_name(ein, spec_name="softmax", name="soft")
    added = Add(ein, soft, name="fin")

    b_prime = Scalar(1.7, shape=(2, 3), name="b_prime")

    def check(matcher: Union[IterativeMatcher, Matcher], expected: Circuit, *updater_args):
        assert matcher.update(added, *updater_args) == expected
        assert added.update(matcher, *updater_args) == expected
        assert Updater(*updater_args)(added, matcher) == expected
        assert Updater(*updater_args).bind(matcher)(added) == expected
        assert matcher.updater(*updater_args)(added) == expected

    a = Scalar(1.0, shape=(1, 2), name="a")
    ein_full_p = Einsum.from_einsum_string("ij,jk->ik", a, b_prime, name="ein")
    soft_full_p = GeneralFunction.new_by_name(ein_full_p, spec_name="softmax", name="soft")
    added_full_p = Add(ein_full_p, soft_full_p, name="fin")

    check(Matcher(b), added_full_p, lambda _: b_prime)

    ein_prime = Einsum.from_einsum_string("ij,jk->ik", a, b_prime, name="ein")
    soft_prime = GeneralFunction.new_by_name(ein_prime, spec_name="softmax", name="soft")
    added_prime = Add(ein, soft_prime, name="fin")

    check(IterativeMatcher.new_chain("soft", "b"), added_prime, lambda _: b_prime)


def test_transform_ident():
    a = Scalar(1.0, shape=(1, 2), name="a")
    assert Transform.ident()(a) == a


def test_get():
    a = Scalar(1.0, shape=(1, 2), name="a")
    b = Scalar(2.0, shape=(2, 3), name="b")
    ein = Einsum.from_einsum_string("ij,jk->ik", a, b, name="ein")
    soft = GeneralFunction.new_by_name(ein, spec_name="softmax", name="soft")
    added = Add(ein, soft, name="fin")
    other_circ = Symbol.new_with_random_uuid((3,))

    assert Getter()(added, Scalar, fancy_validate=True) == {a, b}
    assert Getter()(added, ein, fancy_validate=True) == {ein}
    assert Getter().get_unique_op(added, ein, fancy_validate=True) == ein
    assert Getter().get_unique(added, ein, fancy_validate=True) == ein
    assert Getter().get_unique_op(added, "sdkfj") is None
    assert Getter()(added, "sdkfj") == set()

    with pytest.raises(RuntimeError, match="found no"):
        Getter().get_unique(added, "sdkfj")

    assert Getter()(added, ~Matcher(Scalar), fancy_validate=True) == {ein, soft, added}
    assert (~Matcher(Scalar)).get(added) == {ein, soft, added}
    assert Matcher("a").get_unique(added) == a
    assert Matcher("a").get_unique_op(added) == a
    assert Matcher("sdkfj").get_unique_op(added) is None
    assert Matcher("sdkfj").get(added) == set()

    assert added.get(~Matcher(Scalar), fancy_validate=True) == {ein, soft, added}
    assert added.get(~Matcher(Scalar)) == {ein, soft, added}
    assert added.get_unique("a") == a
    assert added.get_unique(Matcher("a")) == a
    assert added.get_unique(IterativeMatcher(Matcher("a"))) == a
    assert added.get_unique_op("a") == a
    assert added.get_unique_op("sdkfj") is None
    assert added.get("sdkfj") == set()

    assert (~Matcher(Scalar)).getter()(added) == {ein, soft, added}
    assert Matcher("a").getter().get_unique(added) == a
    assert Matcher("a").getter().get_unique_op(added) == a

    with pytest.raises(RuntimeError, match="Didn't match all names"):
        Getter()(added, "sdkfj", fancy_validate=True)
    with pytest.raises(RuntimeError, match="Didn't match all names"):
        Getter()(added, {"sdkfj", "ein"}, fancy_validate=True)
    with pytest.raises(RuntimeError, match="Didn't match all types"):
        Getter()(added, Cumulant, fancy_validate=True)
    with pytest.raises(RuntimeError, match="Didn't match all types"):
        Getter()(added, Matcher.types(Einsum, Cumulant), fancy_validate=True)
    with pytest.raises(RuntimeError, match="Didn't match all types"):
        Matcher.types(Einsum, Cumulant).validate(added)
    with pytest.raises(RuntimeError, match="Didn't match all types"):
        Matcher.types(Einsum, Cumulant).get(added, fancy_validate=True)
    with pytest.raises(RuntimeError, match="Didn't match all types"):
        Matcher.types(Einsum, Cumulant).get_unique_op(added, fancy_validate=True)
    with pytest.raises(RuntimeError, match="Didn't match all types"):
        Matcher.types(Einsum, Cumulant).get_unique(added, fancy_validate=True)
    with pytest.raises(RuntimeError, match="Didn't match all circuits"):
        Getter()(added, other_circ, fancy_validate=True)
    with pytest.raises(RuntimeError, match="Didn't match all circuits"):
        Matcher.circuits(other_circ, a).get(added, fancy_validate=True)
    with pytest.raises(RuntimeError, match="Didn't match all circuit"):
        Matcher.any(Matcher.circuits(other_circ, a), "dif").get(added, fancy_validate=True)
    with pytest.raises(RuntimeError, match="Didn't match all names"):
        Matcher.any("dif", Matcher.circuits(other_circ, a)).get(added, fancy_validate=True)
    with pytest.raises(RuntimeError, match="Didn't match all circuit"):
        Matcher.all(Matcher.circuits(other_circ, a), "dif").get(added, fancy_validate=True)
    with pytest.raises(RuntimeError, match="Didn't match all names"):
        Matcher.all("dif", Matcher.circuits(other_circ, a)).get(added, fancy_validate=True)
    with pytest.raises(RuntimeError, match="Didn't match all names"):
        Matcher.all("dif", Matcher.circuits(other_circ, a)).getter(default_fancy_validate=True).get(
            added, fancy_validate=None
        )
    with pytest.raises(RuntimeError, match="Didn't match all names"):
        Matcher.all("dif", Matcher.circuits(other_circ, a)).getter(default_fancy_validate=False).get(
            added, fancy_validate=True
        )

    a = Scalar(1.0, shape=(1, 2), name="a")
    b = Scalar(2.0, shape=(2, 3), name="b")
    b_dup = Scalar(2.0, shape=(3,), name="b")
    ein = Einsum.from_einsum_string("ij,jk->ik", a, b, name="ein")
    soft = GeneralFunction.new_by_name(ein, spec_name="softmax", name="soft")
    added = Add(ein, soft, b_dup, name="fin")

    assert Matcher("ein").chain(Scalar).get(added, fancy_validate=True) == {a, b}
    assert Matcher("ein").chain("b").get(added, fancy_validate=True) == {b}
    assert Matcher("b").get(added, fancy_validate=True) == {b, b_dup}
    assert Matcher("ein").chain(b_dup).get(added, fancy_validate=False) == set()
    with pytest.raises(RuntimeError, match="Didn't match all circuits"):
        Matcher("ein").chain(b_dup).get(added, fancy_validate=True)
    assert Matcher("ein").chain(a).new_or(Matcher("fin").chain(b_dup)).get(added, fancy_validate=True) == {a, b_dup}

    with pytest.raises(RuntimeError, match="Didn't match all circuits"):
        Matcher("ein").chain(other_circ).new_or(Matcher("fin").chain(b_dup)).get(added, fancy_validate=True)
    with pytest.raises(RuntimeError, match="Didn't match all circuits"):
        Matcher("ein").chain(Matcher.circuits(other_circ, a)).new_or(Matcher("fin").chain(b_dup)).get(
            added, fancy_validate=True
        )
    with pytest.raises(RuntimeError, match="Didn't match all circuits"):
        Matcher("ein").chain(a).new_or(Matcher("fin").chain(Matcher.circuits(other_circ, b_dup))).get(
            added, fancy_validate=True
        )
    with pytest.raises(RuntimeError, match="Didn't match all circuits"):
        Matcher("ein").chain_many([a], [b_dup]).get(added, fancy_validate=True)


def test_any_found():
    a = Scalar(1.0, shape=(1, 2), name="a")
    b = Scalar(2.0, shape=(2, 3), name="b")
    ein = Einsum.from_einsum_string("ij,jk->ik", a, b, name="ein")
    soft = GeneralFunction.new_by_name(ein, spec_name="softmax", name="soft")
    added = Add(ein, soft, name="fin")
    other_circ = Symbol.new_with_random_uuid((3,))

    assert AnyFound()(added, Scalar)
    assert not AnyFound()(added, other_circ)
    assert AnyFound()(added, Einsum)
    assert not AnyFound()(b, Einsum)
    assert not IterativeMatcher(Einsum).any_found()(b)
    assert IterativeMatcher(Einsum).any_found()(added)
    assert not IterativeMatcher(Einsum).are_any_found(b)
    assert IterativeMatcher(Einsum).are_any_found(added)

    # we could add more tests I guess


def test_get_per_child():
    a = Scalar(1.0, shape=(1, 2), name="a")
    b = Scalar(2.0, shape=(2, 3), name="b")
    ein = Einsum.from_einsum_string("ij,jk->ik", a, b, name="ein")

    def run_func(x: Circuit):
        assert x == ein
        return IterateMatchResults([FINISHED, True], False)

    matcher = IterativeMatcher.new_func(run_func)
    res = matcher.match_iterate(ein)
    assert not res.found
    assert isinstance(res.updated, list)
    assert res.updated[0] == FINISHED

    assert Getter()(ein, matcher, fancy_validate=True) == {b}

    def run_func_both(x: Circuit):
        assert x == ein
        return IterateMatchResults([True, True], False)

    matcher = IterativeMatcher.new_func(run_func_both)
    assert Getter()(ein, matcher, fancy_validate=True) == {a, b}

    def matcher_for_both(_: Circuit):
        return IterateMatchResults(["b", "a"], False)

    matcher = IterativeMatcher.new_func(matcher_for_both)
    assert Getter()(ein, matcher, fancy_validate=True) == set()

    ein_new = Einsum.from_einsum_string(
        "ij,jk->ik", Add(b, a.sum()), Scalar(34.0, shape=(3, 7), name="!@*(#&$"), name="ein"
    )
    assert Getter()(ein_new, matcher, fancy_validate=True) == {b}

    ein_new = Einsum.from_einsum_string(
        "jk,ij->ik", Scalar(34.0, shape=(3, 7), name="!@*(#&$"), Add(b, a.sum(name="sum")), name="ein"
    )
    assert Getter()(ein_new, matcher, fancy_validate=True) == {a}


def test_any_found_per_child():
    # literally just copy paste + minor changes to above test
    # could do property based testing with 'get', but I'm not bothering for now
    a = Scalar(1.0, shape=(1, 2), name="a")
    b = Scalar(2.0, shape=(2, 3), name="b")
    ein = Einsum.from_einsum_string("ij,jk->ik", a, b, name="ein")

    def run_func(x: Circuit):
        assert x == ein
        return IterateMatchResults([FINISHED, True], False)

    matcher = IterativeMatcher.new_func(run_func)
    assert AnyFound()(ein, matcher)

    def run_func_both(x: Circuit):
        assert x == ein
        return IterateMatchResults([True, True], False)

    matcher = IterativeMatcher.new_func(run_func_both)
    assert AnyFound()(ein, matcher)

    def matcher_for_both(_: Circuit):
        return IterateMatchResults(["b", "a"], False)

    matcher = IterativeMatcher.new_func(matcher_for_both)
    assert not AnyFound()(ein, matcher)

    ein_new = Einsum.from_einsum_string(
        "ij,jk->ik", Add(b, a.sum()), Scalar(34.0, shape=(3, 7), name="!@*(#&$"), name="ein"
    )
    assert AnyFound()(ein_new, matcher)

    ein_new = Einsum.from_einsum_string(
        "jk,ij->ik", Scalar(34.0, shape=(3, 7), name="!@*(#&$"), Add(b, a.sum(name="sum")), name="ein"
    )
    assert AnyFound()(ein_new, matcher)


def test_update_per_child():
    a = Scalar(1.0, shape=(1, 2), name="a")
    b = Scalar(2.0, shape=(2, 3), name="b")

    def matcher_for_both(_: Circuit):
        return IterateMatchResults(["b", "a"], False)

    matcher = IterativeMatcher.new_func(matcher_for_both)

    ein_new = Einsum.from_einsum_string(
        "ij,jk->ik", Add(b, a.sum(), name="add@#$"), Scalar(34.0, shape=(3, 7), name="!@*(#&$"), name="ein"
    )

    assert Updater(lambda x: Scalar(0.0, shape=x.shape, name="zero"))(
        ein_new, matcher, fancy_validate=True
    ) == Einsum.from_einsum_string(
        "ij,jk->ik",
        Add(Scalar(0.0, b.shape, name="zero"), a.sum(), name="add@#$"),
        Scalar(34.0, shape=(3, 7), name="!@*(#&$"),
        name="ein",
    )

    ein_new = Einsum.from_einsum_string(
        "jk,ij->ik", Scalar(34.0, shape=(3, 7), name="!@*(#&$"), Add(b, a.sum(name="sum"), name="hi"), name="ein"
    )
    assert Updater(lambda x: Scalar(0.0, shape=x.shape, name="zero"))(
        ein_new, matcher, fancy_validate=True
    ) == Einsum.from_einsum_string(
        "jk,ij->ik",
        Scalar(34.0, shape=(3, 7), name="!@*(#&$"),
        Add(b, Scalar(0.0, shape=a.shape, name="zero").sum(name="sum"), name="hi"),
        name="ein",
    )


def test_updater_assert_any_found():
    a = Scalar(1.0, shape=(1, 2), name="a")

    to_zero = lambda x: Scalar(0.0, shape=x.shape, name="zero")

    assert a.update("a", to_zero, assert_any_found=True) == to_zero(a)
    assert a.update({"a", "b"}, to_zero, assert_any_found=True) == to_zero(a)
    assert a.update("b", to_zero) == a
    with pytest.raises(RuntimeError, match="No matches found"):
        a.update("b", to_zero, assert_any_found=True)

    assert IterativeMatcher("a").update(a, to_zero, assert_any_found=True) == to_zero(a)
    assert IterativeMatcher({"a", "b"}).update(a, to_zero, assert_any_found=True) == to_zero(a)
    assert IterativeMatcher("b").update(a, to_zero) == a
    with pytest.raises(RuntimeError, match="No matches found"):
        IterativeMatcher("b").update(a, to_zero, assert_any_found=True)

    assert IterativeMatcher("a").updater(to_zero, default_assert_any_found=True)(a) == to_zero(a)
    assert IterativeMatcher({"a", "b"}).updater(to_zero, default_assert_any_found=True)(a) == to_zero(a)
    assert IterativeMatcher("b").updater(to_zero)(a) == a
    with pytest.raises(RuntimeError, match="No matches found"):
        IterativeMatcher("b").updater(to_zero, default_assert_any_found=True)(a)

    assert IterativeMatcher("a").updater(to_zero)(a, assert_any_found=True) == to_zero(a)
    assert IterativeMatcher({"a", "b"}).updater(to_zero)(a, assert_any_found=True) == to_zero(a)
    assert IterativeMatcher("b").updater(to_zero)(a) == a
    with pytest.raises(RuntimeError, match="No matches found"):
        IterativeMatcher("b").updater(to_zero)(a, assert_any_found=True)

    assert IterativeMatcher("a").updater(to_zero, default_assert_any_found=True).update(a) == to_zero(a)
    assert IterativeMatcher({"a", "b"}).updater(to_zero, default_assert_any_found=True).update(a) == to_zero(a)
    assert IterativeMatcher("b").updater(to_zero).update(a) == a
    with pytest.raises(RuntimeError, match="No matches found"):
        IterativeMatcher("b").updater(to_zero, default_assert_any_found=True).update(a)

    assert IterativeMatcher("a").updater(to_zero).update(a, assert_any_found=True) == to_zero(a)
    assert IterativeMatcher({"a", "b"}).updater(to_zero).update(a, assert_any_found=True) == to_zero(a)
    assert IterativeMatcher("b").updater(to_zero).update(a) == a
    with pytest.raises(RuntimeError, match="No matches found"):
        IterativeMatcher("b").updater(to_zero).update(a, assert_any_found=True)

    lax_updater = Updater(to_zero)
    # loop to test caching behavior
    for _ in range(2):
        assert lax_updater(a, "a", assert_any_found=True) == to_zero(a)
        assert lax_updater(a, {"a", "b"}, assert_any_found=True) == to_zero(a)
        assert lax_updater(a, "b") == a
        with pytest.raises(RuntimeError, match="No matches found"):
            lax_updater(a, "b", assert_any_found=True)
        with pytest.raises(RuntimeError, match="No matches found"):
            lax_updater.update(a, "b", assert_any_found=True)

    strict_updater = Updater(to_zero, default_assert_any_found=True)
    for _ in range(2):
        assert strict_updater(a, "a") == to_zero(a)
        assert strict_updater(a, {"a", "b"}) == to_zero(a)
        assert strict_updater(a, "b", assert_any_found=False) == a
        with pytest.raises(RuntimeError, match="No matches found"):
            strict_updater(a, "b")
        with pytest.raises(RuntimeError, match="No matches found"):
            strict_updater.update(a, "b")


def test_expander_assert_any_found():
    a = Scalar(1.0, shape=(3, 2), name="a")

    to_zero_expand = lambda x: Scalar(0.0, shape=(5, 3, 2), name="zero")

    e_a = Expander(("a", to_zero_expand), default_assert_any_found=True)
    e_a_b_1 = Expander(({"a", "b"}, to_zero_expand), default_assert_any_found=True)
    e_a_b_2 = Expander(("a", to_zero_expand), ("b", to_zero_expand), default_assert_any_found=True)
    e_b_lax = Expander(("b", to_zero_expand))
    e_b_strict = Expander(("b", to_zero_expand), default_assert_any_found=True)

    # loop to test caching behavior
    for _ in range(2):
        assert e_a(a) == to_zero_expand(a)
        assert e_a_b_1(a) == to_zero_expand(a)
        assert e_a_b_2(a) == to_zero_expand(a)
        assert e_b_lax(a) == a
        with pytest.raises(RuntimeError, match="No matches found"):
            e_b_strict(a)

    e_a = Expander(("a", to_zero_expand))
    e_a_b_1 = Expander(({"a", "b"}, to_zero_expand))
    e_a_b_2 = Expander(("a", to_zero_expand), ("b", to_zero_expand))
    e_b = Expander(("b", to_zero_expand))

    # loop to test caching behavior
    for _ in range(2):
        assert e_a(a, assert_any_found=True) == to_zero_expand(a)
        assert e_a_b_1(a, assert_any_found=True) == to_zero_expand(a)
        assert e_a_b_2(a, assert_any_found=True) == to_zero_expand(a)
        assert e_b(a) == a
        with pytest.raises(RuntimeError, match="No matches found"):
            e_b(a, assert_any_found=True)


def test_batch():
    a = Scalar(1.0, shape=(1, 2), name="a")
    b = Scalar(2.0, shape=(2, 3), name="b")
    ein = Einsum.from_einsum_string("ij,jk->ik", a, b, name="ein")
    soft = GeneralFunction.new_by_name(ein, spec_name="softmax", name="soft")
    added = Add(ein, soft, name="fin")

    scalar_batch = lambda x: x.cast_scalar().evolve_shape((5, *x.shape))

    batched = Expander((a, scalar_batch))(added)
    # batched.print()
    assert batched.shape == (5,) + added.shape
    assert batched.name == added.name

    batched = Expander((Matcher.circuits(a, b), scalar_batch))(added)
    # batched.print()
    assert batched.shape == (5,) + added.shape
    assert batched.name == added.name
    canon_ein = Matcher(Einsum).get_unique(batched).cast_einsum().normalize_ints()
    assert canon_ein.all_input_axes() == [(0, 1, 2), (0, 2, 3)]
    assert canon_ein.out_axes == (0, 1, 3)
    assert canon_ein.name == ein.name

    batched = Expander((a, scalar_batch), (b, scalar_batch))(added)
    # batched.print()
    assert batched.shape == (5,) + added.shape
    assert batched.name == added.name
    canon_ein = Matcher(Einsum).get_unique(batched).cast_einsum().normalize_ints()
    assert canon_ein.all_input_axes() == [(0, 1, 2), (0, 2, 3)]
    assert canon_ein.out_axes == (0, 1, 3)
    assert canon_ein.name == ein.name

    batched = Expander((a, scalar_batch), (b, scalar_batch), suffix="batch_suffix")(added)
    # batched.print()
    assert batched.shape == (5,) + added.shape
    assert batched.name == added.name + "_batch_suffix"
    canon_ein = Matcher(Einsum).get_unique(batched).cast_einsum().normalize_ints()
    assert canon_ein.all_input_axes() == [(0, 1, 2), (0, 2, 3)]
    assert canon_ein.out_axes == (0, 1, 3)
    assert canon_ein.name == ein.name + "_batch_suffix"


def test_restrict_module_spec():
    args = {"ln.w.bias": Array.randn(17), "ln.w.scale": Array.randn(17), "ln.input": Array.randn(17)}
    mod = Module(layernorm.spec, "ln_out", **args)
    assert new_traversal(start_depth=1).filter_module_spec().get(mod) == set(args.values())

    mod = module_new_bind(
        TransformerBlockParams().get().body,
        ("a.ln.w.bias", Array.randn(17)),
        ("a.ln.w.scale", Array.randn(17)),
        ("b.input", Array.randn(5, 17)),
    )
    all_a_inputs = (
        Matcher("a.norm")
        .chain(new_traversal(start_depth=1).filter_module_spec())
        .get(TransformerBlockParams().get().body)
    )
    assert all_a_inputs == Matcher({"a.norm.input", "a.ln.w.scale", "a.ln.w.bias"}).get(mod)

    all_mlp_sub_inputs = (
        Matcher("m.norm_call")
        .chain(restrict_sl(Matcher(Module).new_not(), depth_slice=S[1:]).filter_module_spec())
        .get(TransformerBlockParams().get().body)
    )
    assert all_mlp_sub_inputs == Matcher({"m.norm.input", "m.ln.w.bias", "m.ln.w.scale"}).get(mod)

    all_mlp_sub_inputs = (
        Matcher("m.norm_call")
        .chain(restrict_sl(Matcher(Module).new_not(), depth_slice=S[1:]).filter_module_spec())
        .get(mod)
    )
    assert all_mlp_sub_inputs == Matcher({"m.norm.input", "m.ln.w.bias", "m.ln.w.scale"}).get(mod)

    scale_val = Array.randn(17, name="arr_scale").rearrange_str("a->a", name="perm")
    mod_expand = module_new_bind(
        TransformerBlockParams().get().body,
        ("m.ln.w.scale", scale_val),
    ).substitute()

    all_mlp_sub_inputs = (
        Matcher("m.norm_call")
        .chain(restrict_sl(Matcher(Module).new_not(), depth_slice=S[1:]).filter_module_spec())
        .get(mod_expand)
    )
    assert all_mlp_sub_inputs == Matcher({"m.norm.input", "m.ln.w.bias"}).get(mod).union(all_children(scale_val))


def test_spec_circuit_matcher():
    mod = module_new_bind(
        TransformerBlockParams().get().body,
        ("a.ln.w.bias", Array.randn(17)),
        ("a.ln.w.scale", Array.randn(17)),
        ("b.input", Array.randn(5, 17)),
    )

    print([x.name for x in IterativeMatcher(Module).spec_circuit_matcher().get(mod)])
    assert IterativeMatcher(Module).spec_circuit_matcher().get(mod) == Matcher(
        {"b", "m.norm_call", "m", "ln", "a.norm_call", "not_mask", "a", "a.head.on_inp", "a.head"}
    ).get(mod, fancy_validate=True)

    assert IterativeMatcher("m.norm").spec_circuit_matcher().get(mod) == Matcher({"ln"}).get(mod)


def test_children_matcher():
    mod = module_new_bind(
        TransformerBlockParams().get().body,
        ("a.ln.w.bias", Array.randn(17)),
        ("a.ln.w.scale", Array.randn(17)),
        ("b.input", Array.randn(5, 17)),
    )

    print(mod.get_unique("m.norm").children)

    assert IterativeMatcher("m.norm").children_matcher({0}).get(mod) == Matcher({"ln"}).get(mod)

    assert IterativeMatcher("m.norm").children_matcher({0, 1}).get(mod) == Matcher({"ln", "m.norm.input"}).get(mod)
    assert IterativeMatcher("m.norm").children_matcher({0, 3}).get(mod) == Matcher({"ln", "m.ln.w.scale"}).get(mod)
    assert IterativeMatcher("m.norm").children_matcher({2, 1}).get(mod) == Matcher({"m.norm.input", "m.ln.w.bias"}).get(
        mod
    )

    with pytest.raises(IterativeMatcherChildNumbersOutOfBoundsError):
        IterativeMatcher("m.norm").children_matcher({4}).get(mod)
    with pytest.raises(IterativeMatcherChildNumbersOutOfBoundsError):
        IterativeMatcher("m.norm").children_matcher({7}).get(mod)


def test_module_arg_matcher():
    mod = module_new_bind(
        TransformerBlockParams().get().body,
        ("a.ln.w.bias", Array.randn(17)),
        ("a.ln.w.scale", Array.randn(17)),
        ("b.input", Array.randn(5, 17)),
    )

    PrintOptions().print(mod)

    assert IterativeMatcher("m.norm").module_arg_matcher("ln.input").get(mod) == Matcher({"m.norm.input"}).get(mod)
    assert IterativeMatcher("m.norm").module_arg_matcher("ln.w.bias").get(mod) == Matcher({"m.ln.w.bias"}).get(mod)
    assert IterativeMatcher("m.norm").module_arg_matcher("ln.w.scale").get(mod) == Matcher({"m.ln.w.scale"}).get(mod)
    assert IterativeMatcher("m.norm").module_arg_matcher({"ln.input", "ln.w.bias", "ln.w.scale"}).get(mod) == Matcher(
        {"m.norm.input", "m.ln.w.bias", "m.ln.w.scale"}
    ).get(mod)

    # module arg matchers match by argument *symbol*, not argument *content*, and hence cannot match
    # the module spec (which does not have an argument symbol)
    assert IterativeMatcher("m.norm").module_arg_matcher("ln").get(mod) == set()

    assert IterativeMatcher("a.norm_call").module_arg_matcher("a.input").get(mod) == Matcher("a.norm").get(mod)

    # Check that we only search for arguments under modules which match the first matcher
    assert IterativeMatcher("a.norm_call").module_arg_matcher("ln.input").get(mod) == set()
    assert IterativeMatcher("m.norm").module_arg_matcher("a.input").get(mod) == set()


def test_print_matcher():
    setups = [
        (Matcher("a_name"), '''"a_name"'''),
        (Matcher({"a_name", "b_other"}), """{"a_name", "b_other"}"""),
        (Matcher(set[Type[Circuit]]({Scalar, Einsum})), """{Einsum, Scalar}"""),
        (
            Matcher.all(Matcher(set[Type[Circuit]]({Scalar, Einsum})), Matcher({"a_name", "b_other"})),
            """All({Einsum, Scalar}, {"a_name", "b_other"})""",
        ),
        (
            Matcher.any(Matcher(set[Type[Circuit]]({Scalar, Einsum})), Matcher.new_not(Matcher({"a_name", "b_other"}))),
            """Any({Einsum, Scalar}, Not {"a_name", "b_other"})""",
        ),
        (Matcher.true_matcher(), """Always"""),
        (Matcher.false_matcher(), """Never"""),
        (Matcher(Scalar(1.0, shape=(1, 2), name="a_name")), """'a_name' Scalar 80114e"""),
        (
            Matcher(
                set[Circuit]({Scalar(1.0, shape=(1, 2), name="a_name"), Scalar(2.0, shape=(2, 3), name="b_other")})
            ),
            """{'a_name' Scalar 80114e, 'b_other' Scalar c85657}""",
        ),
        (Matcher.regex("^.$"), """re-escdot'^.$'"""),
        (Matcher.regex("^.$", escape_dot=False), """re'^.$'"""),
    ]
    for matcher, target in setups:
        print(matcher)
        assert repr(matcher) == target
    print(Matcher(lambda x: x.name.startswith("hi")))


def test_print_iterative_matcher():
    setups = [
        (IterativeMatcher.new_chain("a_name", "b_name"), """["a_name" -> "b_name"]"""),
        (
            IterativeMatcher.any(
                IterativeMatcher.new_chain("a_name", "b_name"), IterativeMatcher.new_chain("b_name", "a_name")
            ),
            """Any(["a_name" -> "b_name"], ["b_name" -> "a_name"])""",
        ),
        (IterativeMatcher("a_name"), '''"a_name"'''),
        (IterativeMatcher.new_children_matcher("a_name", frozenset([0, 1])), """Children("a_name", [0,1])"""),
        (IterativeMatcher.new_module_arg_matcher("a_name", "b_name"), """ModuleArg("a_name", "b_name")"""),
        (IterativeMatcher.new_spec_circuit_matcher("a_name"), """SpecCircuit("a_name")"""),
        (IterativeMatcher("a_name").filter_module_spec(), """NoModuleSpec("a_name")"""),
        (
            new_traversal(start_depth=1, end_depth=3),
            """Restrict{depth_range: 1:3}""",
        ),
        (
            restrict(
                restrict(
                    restrict(
                        IterativeMatcher.new_chain(
                            "n;asdkjfn;lasknfl;wkenf;alknklk;lasdknfl;an", "b", "c", "d", "efghijk"
                        ),
                        start_depth=1,
                        end_depth=10,
                    ),
                    term_early_at={"asdfasdfasdf", "1234567890", "12345678912345678901234567890", "qwertyuiop"},
                ),
                term_if_matches=True,
            ),
            """Restrict{
    matcher: Restrict{    
        matcher: Restrict{        
            depth_range: 1:10        
            matcher: ["n;asdkjfn;lasknfl;wkenf;alknklk;lasdknfl;an" -> "b" -> "c" -> "d" -> "efghijk"]        
        }    
        term_at: {        
            "1234567890"        
            "12345678912345678901234567890"        
            "asdfasdfasdf"        
            "qwertyuiop"        
        }    
    }
    term_if_matches: true
}""",
        ),
    ]
    for matcher, target in setups:
        print(matcher)
        assert repr(matcher) == target
    print(IterativeMatcher.any("a_name", lambda x: False))


def test_iterative_matcher_name_append():
    TransformerBlockParams().get().body.print()
    matcher = IterativeMatcher.new_chain("b.m", "m.norm_call", GeneralFunction)
    append_matchers_to_names(TransformerBlockParams().get().body, matcher).print()


def test_matcher_debug():
    matcher = IterativeMatcher.new_chain("b.mlp", "gelu_mlp", GeneralFunction)
    print_matcher_debug(TransformerBlockParams().get().body, matcher, PrintOptions())
    print_matcher_debug(TransformerBlockParams().get().body, matcher, PrintOptions(arrows=True))
    print_matcher_debug(
        rc.Parser()(
            """
'b.mlp' Einsum ,,->
  'gelu_mlp' Einsum ,->
    's0' [] Symbol
    's1' [] Symbol
  'gelu_mlp'
  's0'
    """
        ),
        matcher,
        PrintOptions(arrows=True),
    )


def test_iterative_matcher_and():
    a = Scalar(1.0, shape=(1, 2), name="a_name")
    b = Scalar(2.0, shape=(2, 3), name="b_other")
    c = Einsum.from_einsum_string("ab,bc->ac", a, b, name="c_name")

    def check_for_same(m: IterativeMatcher, c: Circuit, found: bool):
        res = m.match_iterate(c)
        assert res.updated is None
        assert res.found == found

    check_for_same(IterativeMatcher("a_name") & "b_other", a, False)
    check_for_same(IterativeMatcher("a_name") & "b_other", a, False)
    check_for_same(IterativeMatcher("a_name") | "b_other", a, True)
    check_for_same("b_other" & IterativeMatcher("a_name"), a, False)
    check_for_same("b_other" | IterativeMatcher("a_name"), a, True)
    check_for_same(IterativeMatcher.all("a_name", Scalar), a, True)
    check_for_same(IterativeMatcher.all("a_name", Scalar), b, False)

    check_for_same(IterativeMatcher.all("a_name", Scalar), a, True)
    check_for_same(IterativeMatcher.all("a_name", Scalar), b, False)
    check_for_same(IterativeMatcher.all("a_name", Scalar), c, False)

    check_for_same(IterativeMatcher.all("a_name"), a, True)
    check_for_same(IterativeMatcher.all("a_name"), b, False)
    check_for_same(IterativeMatcher.all("a_name"), c, False)

    out = IterativeMatcher.all("hi", restrict("a_name", end_depth=3)).match_iterate(c)
    assert not out.found
    up = restrict("a_name", end_depth=3).match_iterate(c).updated
    assert up is not None
    assert not isinstance(up, list)
    assert not isinstance(up, Finished)
    assert out.updated == IterativeMatcher.all("hi", up)

    out = IterativeMatcher.all(Scalar, restrict("a_name", end_depth=3)).match_iterate(a)
    assert out.found
    up = restrict("a_name", end_depth=3).match_iterate(c).updated
    assert up is not None
    assert not isinstance(up, list)
    assert not isinstance(up, Finished)
    assert out.updated == IterativeMatcher.all(Scalar, up)

    out = IterativeMatcher.all(Einsum, restrict("a_name", end_depth=3)).match_iterate(a)
    assert not out.found
    up = restrict("a_name", end_depth=3).match_iterate(c).updated
    assert up is not None
    assert not isinstance(up, list)
    assert not isinstance(up, Finished)
    assert out.updated == IterativeMatcher.all(Einsum, up)

    out = IterativeMatcher.all(Einsum, IterativeMatcher.term()).match_iterate(a)
    assert not out.found
    assert out.updated == FINISHED

    out = IterativeMatcher.all(Scalar, IterativeMatcher.term()).match_iterate(a)
    assert not out.found
    assert out.updated == FINISHED

    out = IterativeMatcher.all(IterativeMatcher.term(), Scalar).match_iterate(a)
    assert not out.found
    assert out.updated == FINISHED

    out = IterativeMatcher.all(Scalar, IterativeMatcher.term(match_next=True)).match_iterate(a)
    assert out.found
    assert out.updated == FINISHED

    out = IterativeMatcher.all(Scalar, restrict("c_name", end_depth=3).children_matcher({1})).match_iterate(c)
    assert not out.found
    up = restrict("c_name", end_depth=3).children_matcher({1}).match_iterate(c).updated
    assert up is not None
    assert isinstance(up, list)
    new_up: List[IterativeMatcher] = []
    for u in up:
        assert not isinstance(u, Finished)
        new_up.append(u)

    assert out.updated == [IterativeMatcher.all(Scalar, new_up[0]), IterativeMatcher.all(Scalar, new_up[1])]

    out = IterativeMatcher.all(
        Scalar, IterativeMatcher.new_func(lambda _: IterateMatchResults([FINISHED, IterativeMatcher(True)], found=True))
    ).match_iterate(c)
    assert not out.found

    assert out.updated == [FINISHED, IterativeMatcher.all(Scalar, IterativeMatcher(True))]

    with pytest.raises(TypeError, match="coerced"):
        IterativeMatcher("a_name") and IterativeMatcher("b_other")


def test_get_paths():
    a = Scalar(1.0, name="match")
    a_bis = Add(name="other")
    b = Add(a, a_bis, name="parent")
    c = Add(b, name="grandparent")

    expected = {a: [a, b, c]}
    assert Getter().get_paths(c, Matcher(Scalar)) == expected
    assert Matcher(Scalar).get_paths(c) == expected
    assert c.get_paths(Matcher(Scalar)) == expected
    assert Getter().bind(Scalar).get_paths(c) == expected

    assert Getter().get_paths(c, Matcher(Array)) == {}

    assert Getter().get_paths(c, {a, c, a_bis}) == {a: [a, b, c], a_bis: [a_bis, b, c], c: [c]}


def test_get_all_paths():
    a = Scalar(1.0, name="match")
    a_bis = Add(name="other")
    b = Add(a, a_bis, name="parent")
    c = Add(b, name="grandparent")

    expected = {a: [[a, b, c]]}
    assert Getter().get_all_paths(c, Matcher(Scalar)) == expected
    assert Matcher(Scalar).get_all_paths(c) == expected
    assert c.get_all_paths(Matcher(Scalar)) == expected
    assert Getter().bind(Scalar).get_all_paths(c) == expected

    assert Getter().get_all_paths(c, Matcher(Array)) == {}

    assert Getter().get_all_paths(c, {a, c, a_bis}) == {a: [[a, b, c]], a_bis: [[a_bis, b, c]], c: [[c]]}

    d = Add(a, a)
    assert Getter().get_all_paths(d, Matcher(Scalar)) == {a: [[a, d], [a, d]]}

    e = Add(c, b)
    r = Getter().get_all_paths(e, Matcher(Scalar))
    assert r == {a: [[a, b, c, e], [a, b, e]]} or r == {a: [[a, b, e], [a, b, c, e]]}


def test_non_callable_errors_on_construct():
    class RandomObj:
        ...

    obj = RandomObj()

    with pytest.raises(TypeError):
        Matcher(obj)  # type: ignore
    with pytest.raises(TypeError):
        IterativeMatcher(obj)  # type: ignore
    with pytest.raises(TypeError):
        IterativeMatcher.new_func(obj)  # type: ignore



from typing import Callable, Tuple

import torch

from rust_circuit import Array
from rust_circuit.causal_scrubbing.dataset import Dataset


def loss_fn(ds: Dataset, out_dict) -> float:
    if isinstance(out_dict, dict):
        out = list(out_dict.values())[0]
    else:
        out = out_dict
    out = out.to(torch.float)
    return torch.abs(out.cpu() - ds.labels.value).mean().item()


class IntDataset(Dataset):
    inp_tensor: torch.Tensor

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        object.__setattr__(self, "inp_tensor", self.xs.value)

    @classmethod
    def of_shape(cls, shape: Tuple, labeler: Callable[[torch.Tensor], torch.Tensor], maxint: int = 10):
        xs = torch.randint(high=maxint, size=shape)
        xs_arr = Array(xs, "xs")
        labels_arr = Array(torch.stack([labeler(x) for x in xs]), "labels")
        ds = cls({"xs": xs_arr, "labels": labels_arr})
        return ds

    def __str__(self):
        if len(self) > 1:
            return f"IntDataset(xs.shape={self.inp_tensor.shape})"
        else:
            xs = self.inp_tensor.item() if self.inp_tensor.numel() == 1 else self.inp_tensor[0].tolist()
            return f"IntDatum(xs={xs}, label={self.labels.value.item()})"



def test_rust_alias_imports():
    from rust_circuit import (
        Axis,
        Binder,
        CircuitColorer,
        CircuitHtmlColorer,
        CliColor,
        GeneralFunctionSpec,
        IntOrMatcher,
        IrreducibleNode,
        IterativeMatcherIn,
        Leaf,
        LeafConstant,
        MatcherIn,
        MaybeUpdate,
        ModuleConstructCallback,
        NestAddsSpec,
        NestAddsSpecMultiple,
        NestAddsSpecSub,
        NestedModuleNamer,
        NestEinsumsSpec,
        NestEinsumsSpecMultiple,
        NestEinsumsSpecSub,
        PrintOptionsBase,
        SampleSpecIn,
        Shape,
        TorchAxisIndex,
        TransformIn,
        UpdatedIterativeMatcher,
        UpdatedIterativeMatcherIn,
        Var,
    )

    _ = (
        Shape,
        Axis,
        IrreducibleNode,
        Leaf,
        LeafConstant,
        Var,
        MatcherIn,
        IterativeMatcherIn,
        TransformIn,
        SampleSpecIn,
        TorchAxisIndex,
        IntOrMatcher,
        NestEinsumsSpecMultiple,
        NestEinsumsSpecSub,
        NestEinsumsSpec,
        NestAddsSpecMultiple,
        NestAddsSpecSub,
        NestAddsSpec,
        GeneralFunctionSpec,
        Binder,
        UpdatedIterativeMatcher,
        UpdatedIterativeMatcherIn,
        PrintOptionsBase,
        CliColor,
        CircuitColorer,
        CircuitHtmlColorer,
        NestedModuleNamer,
        ModuleConstructCallback,
        MaybeUpdate,
    )



from __future__ import annotations

from typing import Dict, Literal, Optional, Union
from warnings import warn

import attrs
import torch

import rust_circuit as rc
import rust_circuit.optional as op

from .dataset import Dataset, color_dataset
from .hypothesis import Correspondence, IllegalCorrespondenceError, InterpNode, SampledInputs, to_inputs

ExperimentCheck = Union[bool, Literal["fast"]]


class ExperimentEvalSettings:
    """
    Settings for ScrubbedExperiment.evaluate.

    optimize: whether to optimize before evaluating. If true, evaluate uses less memory and more time.
    optim_settings: settings for optimizer. Only used if optimize=True.
    batch_size: if set, reshapes circuit into batches before evaluating, to take less memory and more time.
    run_on_all: if True, runs the scrubbed model on the whole batch of scrubbed inputs in order; see
        rc.DiscreteVarAllSpec for more info. likely you want this to be True. if True, num_samples_should not be set.
    num_samples: number of samples to sample when using rc.RandomSampleSpec, so should not be set if run_on_all=True.
        if run_on_all=False and this is not set, will default to the size of the group passed to get_sampler. likely
        you don't want to set this: it only affects *evaluation of the scrubbed circuit*, and not the scrubbing process.
    device_dtype: device and dtype to evaluate on. if only a str is supplied, this is assumed to be the device.
    """

    optimize: bool
    optim_settings: rc.OptimizationSettings
    batch_size: Optional[int]
    run_on_all: bool  # set to False to sample from scrubbed inputs, else will run on all of them in order
    num_samples: Optional[int]
    device_dtype: rc.TorchDeviceDtypeOp

    def __init__(
        self,
        optimize=True,
        optim_settings=rc.OptimizationSettings(),
        batch_size=None,
        run_on_all=True,
        num_samples=None,
        device_dtype: Union[rc.TorchDeviceDtypeOp, str] = rc.TorchDeviceDtypeOp("cuda:0"),
    ):
        if run_on_all and (num_samples is not None):
            raise ValueError(f"num_samples should be None when run_on_all is True, but was {num_samples}")
        self.optimize = optimize
        self.run_on_all = run_on_all
        self.num_samples = num_samples
        self.batch_size = batch_size
        if isinstance(device_dtype, str):
            device_dtype = rc.TorchDeviceDtypeOp(device_dtype, None)
        self.device_dtype = device_dtype
        if self.optimize:
            self.optim_settings = optim_settings

    def get_sampler(self, group, seed):
        """
        First, creates a sample spec: if run_on_all, this is just a RunDiscreteVarAllSpec targeting the specified
        group; otherwise, a RandomSampleSpec that sets a seed for this group. Note seeds are not set for other groups:
        we don't explicitly support having more than one group.

        Then, creates a sampler that will substitute modules, batch, and cast the circuit after sampling.
        """
        sample_spec: rc.SampleSpecIn
        if self.run_on_all:
            sample_spec = rc.RunDiscreteVarAllSpec([group])
        else:
            sample_spec = rc.RandomSampleSpec(
                seeder=lambda g: seed if g == group else None,
                shape=(op.unwrap_or(self.num_samples, len(group)),),
                probs_and_group_evaluation_settings=rc.OptimizationSettings(),
            )

        maybe_batch = (
            lambda x: x
            if self.batch_size is None
            else rc.batch_to_concat(
                x,
                axis=0,
                batch_size=self.batch_size,
            )
        )

        return rc.Sampler(
            sample_spec,
            run_on_sampled=lambda x: rc.cast_circuit(maybe_batch(rc.substitute_all_modules(x)), self.device_dtype),
        )


class Experiment:
    """
    A causal scrubbing experiment.

    The circuit you pass in can be the *complete* circuit to scrub--that is, computing the full behavior function you want
    the scrubbed expectation of, or it can be a subset. For example, this can be convenient if you want to pass in a circuit
    representing a neural network, and compute its scrubbed loss yourself.

    circuit: circuit to be scrubbed. assumes it has inputs matching dataset.input_names with no batch dimension.
    dataset: dataset to both use as reference and sample from for scrubbing. that is, we run causal scrubbing on a batch.
    corr: hypothesized correspondence.
    random_seed: seed for RNG for replicability. will be used 1) to set up a torch generator that is consumed every time we
        need to sample inputs, and 2) for eval if you use ExperimentEvalSettings.run_on_all = False.
    check: whether to check the correspondence for validity. recommend setting it to True, switching to "fast" if your code
        is frustratingly slow; and back to True if you are having confusing errors or results.
    """

    def __init__(
        self,
        circuit: rc.Circuit,
        dataset: Dataset,
        corr: Correspondence,
        random_seed=42,
        check: ExperimentCheck = True,
    ):
        self._base_circuit: rc.Circuit = circuit
        self._dataset = dataset
        self._check = check

        if self._check is not True:
            warn(f"You're using check = {self._check}, your correspondence may have issues that won't be caught!")
        if self._check:
            corr.check(self._base_circuit, circuit_treeified=False)

        self._rng = torch.Generator(device=circuit.device)
        self._rng.manual_seed(random_seed)
        self._seed = random_seed
        self._nodes = corr

    @property
    def base_circuit(self) -> rc.Circuit:
        return self._base_circuit

    @property
    def dataset(self) -> Dataset:
        return self._dataset

    @property
    def nodes(self) -> Correspondence:
        return self._nodes

    def make_ref_ds(self, num_examples: int):
        return self._dataset.sample(num_examples, self._rng)

    def scrub(
        self,
        ref_ds_or_num: Union[Dataset, int],
        group: Optional[rc.Circuit] = None,
        treeify: bool = True,
        check: ExperimentCheck = True,
    ) -> ScrubbedExperiment:
        """
        Performs the _project_ operation of causal scrubbing, except for the actual evaluation.
        Call .evaluate() on the returned ScrubbedExperiment to do that.

        ref_ds: can be used to set the reference dataset to a particular value. If an int is specified, randomly samples
            that size reference dataset from self.dataset, which is normally what you should do when causal scrubbing.
            You might set ref_ds if you want to examine a particular subset of your dataset, however.
        group: the probs_and_group to use for wrapping your inputs in var, and for sampling. Note that the probs
            are ignored if you end up evaluating with run_on_all = True. Must match len(ref_ds) aka num_examples.
        check: whether to check the correspondence for validity
        treeify: will perform an explicit treeification and extra checks
        """
        check = bool(self._check if check is None else check)
        if isinstance(ref_ds_or_num, int):
            num_examples = ref_ds_or_num
            ref_ds = self.make_ref_ds(num_examples)
        else:
            ref_ds = ref_ds_or_num
            num_examples = len(ref_ds)

        if group is None:
            group = rc.DiscreteVar.uniform_probs_and_group(num_examples)
        else:
            if group.shape != (num_examples,):
                raise ValueError(f"group has shape {group.shape=}, should have shape={(num_examples,)}")

        sampled_inputs = self.sample(ref_ds=ref_ds)
        if treeify:
            circuit = self.treeified(check)
            if check:
                self._nodes.check_injective_on_treeified(circuit)
        else:
            circuit = self._base_circuit
        circuit = self.wrap_in_var(circuit, ref_ds, group)
        circuit = self.replace_inputs(circuit, sampled_inputs)
        return ScrubbedExperiment(
            circuit, ref_ds=ref_ds, sampled_inputs=sampled_inputs, group=group, nodes=self._nodes, seed=self._seed
        )

    def sample(self, ref_ds: Dataset) -> SampledInputs:
        """Samples from the interpretation with the given reference dataset"""
        return self._nodes.sample(self._dataset, self._rng, ref_ds)

    def treeified(self, check: ExperimentCheck = True):
        """
        Treeifies the circuit as much as is required by the correspondence, and checks corresponcence
        validity along the way.

        Note this does not depend at all on the interpretation, only on what parts of the circuit
        are mapped! and may be moved into ModelBranch in the future.

        This step is optional: it can be nice to examine your treeified model without setting
        scrubbed inputs, if you skip it then setting scrubbed inputs will treeify the model
        as a side effect.

        This is an algrebraic rewrite!
        """

        def replacement(i_node: InterpNode, a: rc.Circuit) -> rc.Tag:
            input_name = a.name

            if i_node != self._nodes.get_root():
                # Not the root, so inputs upstream of this should have already been tagged
                a = a.cast_tag().node
                if check:  # maybe don't run if check is "fast"? unsure whether this is slow.
                    assert f"{input_name}_scrub_" in a.name, (
                        "We've already seen the root but haven't replaced all inputs, should never happen!",
                        i_node.name,
                        a.name,
                        input_name,
                    )
                    # What node tagged it? Should be a parent of this node
                    prev_tagger_name = a.name.replace(f"{input_name}_scrub_", "")
                    prev_tagger = self._nodes.get_by_name(prev_tagger_name)
                    if i_node not in prev_tagger.children:
                        raise IllegalCorrespondenceError(i_node, prev_tagger)
                    # Check that this parent-child pair respects digraph structure: we know now that the child matches
                    # only paths matched by the parent, but this could also happen if the parent matches nodes that
                    # are *only children* of the nodes matched by the child.
                    parent_circs = self._nodes[prev_tagger].get(circuit)
                    this_circs = self._nodes[i_node].get(circuit)
                    for this_c in this_circs:
                        if not any(parent_c.are_any_found(rc.Matcher(this_c)) for parent_c in parent_circs):
                            raise IllegalCorrespondenceError(
                                "Expected child to be a descendant of at least one parent but was not!",
                                prev_tagger,
                                i_node,
                                this_c,
                            )

            # Note the *input* is being given a unique name, while the outer tag keeps the input_name.
            # This way, when we match on input_name, we'll actually see that there are multiple matches
            # after treeification!
            new_a = a.rename(f"{input_name}_scrub_{i_node.name}")
            return rc.Tag.new_with_random_uuid(new_a, name=input_name)

        circuit = self._base_circuit
        for i_node, m_node in self._nodes.in_dfs_order():
            # we use dataset here, but all that matters is the names
            # (which must match names in the circuit, but don't all have to exist).
            input_matcher = to_inputs(m_node, self._dataset)
            circuit = input_matcher.update(circuit, lambda a: replacement(i_node, a))
        return circuit

    def replace_inputs(self, circuit: rc.Circuit, sampled_inputs: SampledInputs) -> rc.Circuit:
        """
        Replaces inputs to model branches according to the correspondence:
        - conditional sampled dataset at each leaf
        - other-children dataset at internal nodes (relevant iff correspondence is not surjective)

        Works regardless of whether model has already been treeified.

        N.B. This is not an algebraic rewrite!
        """
        for parent_node, interp_node, m in self._nodes.in_dfs_order_with_parents():
            circuit = self._replace_one_input(
                circuit, interp_node, matcher=m, sampled_inputs=sampled_inputs, parent_node=parent_node
            )
        return circuit

    def _replace_one_input(
        self,
        circuit: rc.Circuit,
        interp_node: InterpNode,
        matcher: rc.IterativeMatcher,
        sampled_inputs: SampledInputs,
        parent_node: Optional[InterpNode] = None,
    ) -> rc.Circuit:
        """Internal method"""
        if parent_node is not None and sampled_inputs[parent_node] == sampled_inputs[interp_node]:
            return circuit
        input_matcher = to_inputs(matcher, self._dataset)
        if not input_matcher.are_any_found(circuit):
            raise IllegalCorrespondenceError("No matches found in circuit", input_matcher)

        circuit = input_matcher.update(circuit, lambda a: sampled_inputs[interp_node].arrs[a.name])
        return circuit

    def wrap_in_var(self, circuit: rc.Circuit, ref_ds: Dataset, group: rc.Circuit):
        # TODO: add tests / better support for wrap_in_var = False. Also, add better errors or warnings for input
        # of wrong shape, redundant vars, etc.

        # we use ref_ds here, but we'll overwrite the values later.
        # All that matter are the shape (must match future datasets) and the names
        # (which must match names in the circuit, but don't all have to exist; thus fancy_validate=False).
        def arr_to_var(arr: rc.Circuit):
            arr = ref_ds.arrs[arr.name]
            return rc.DiscreteVar(
                values=arr,
                probs_and_group=group,
                name=f"{arr.name}_var",
            )

        circuit = rc.IterativeMatcher(set(ref_ds.input_names)).update(circuit, arr_to_var, fancy_validate=False)
        return circuit


@attrs.define(frozen=True)
class ScrubbedExperiment:
    """
    The result of calling `Experiment.scrub()`. Keeps track of the scrubbed circuit, the reference dataset, and the
    sampled inputs for all the interpretation nodes.
    """

    circuit: rc.Circuit
    ref_ds: Dataset
    sampled_inputs: SampledInputs
    group: rc.Circuit
    seed: int

    # only needed for printing
    nodes: Correspondence

    def evaluate(self, eval_settings: ExperimentEvalSettings = ExperimentEvalSettings()) -> torch.Tensor:
        """
        Returns the output of the scrubbed circuit, evaluated on the reference dataset.

        If the circuit is a subset of a bigger circuit, you can use .ref_ds to evaluate the other parts of the
        circuit as well.
        """
        sampler = eval_settings.get_sampler(self.group, self.seed)
        circ = sampler.sample(self.circuit)
        return (
            rc.optimize_and_evaluate(
                circ,
                eval_settings.optim_settings,
            )
            if eval_settings.optimize
            else circ.evaluate()
        )

    def print(
        self,
        options: Optional[Union[rc.PrintOptions, rc.PrintHtmlOptions]] = None,
        print_data=True,
        color_by_data=True,
        repr=False,
    ):
        """
        Print the scrubbed circuit, terminating at leaf mapped nodes and at nodes that are not ancestors of any mapped node.
        Optionally annotate all the nodes with the data sampled for them.

        Parameters:
            print_data: if True, prints a sample datum from cond_sampler as a comment
            color_by_data: if True, colors mapped nodes by the data they are sampled from and greys out nodes that are not ancestors of a mapped node
            repr: if True, returns string instead of printing it (used for testing)
        """
        new_options: Union[rc.PrintOptions, rc.PrintHtmlOptions]
        if options is None:
            new_options = rc.PrintHtmlOptions(traversal=rc.new_traversal())
        else:
            new_options = options.evolve()

        # Find all circuits which are mapped to by correspondence and record what dataset they're getting
        scrubbed_nodes: Dict[rc.Circuit, InterpNode] = {
            c: i for i in self.nodes.corr.keys() for c in list(self.nodes[i].get(self.circuit))
        }
        scrubbed_inputs: Dict[rc.Circuit, InterpNode] = {
            c_input: i for c, i in scrubbed_nodes.items() for c_input in c.get(self.ref_ds.input_names)
        }
        scrubbed_matcher = rc.Matcher(set(scrubbed_nodes.keys()))

        # Is it outside the image of the correspondence, i.e. is it not in any path matched by the corr's matchers?
        # note we're not using matcher's get_all_paths because it results in a giant list of nodes; since
        # the model has been treeified, it's sufficient to check whether a node has scrubbed nodes downstream of it.
        not_in_scrubbed_path = rc.Matcher(lambda c: not scrubbed_matcher.are_any_found(c))
        # Expand the children by default?
        scrubbed_leaf_matcher = rc.Matcher(lambda c: scrubbed_matcher(c) and scrubbed_nodes[c].is_leaf())
        term_at_matcher = not_in_scrubbed_path | scrubbed_leaf_matcher

        # Colors!
        def get_color(c: rc.Circuit, html: bool):
            # non html colors are pretty made up, would be nice if printer had a "color by feature"
            if scrubbed_matcher(c):
                return color_dataset(self.sampled_inputs[scrubbed_nodes[c]], html)
            elif c in scrubbed_inputs:
                return color_dataset(self.sampled_inputs[scrubbed_inputs[c]], html)
            elif not_in_scrubbed_path(c):
                return "darkgray" if html else 6  # todo light mode
            else:
                return "lightgray" if html else 13

        if color_by_data:
            new_options.colorer = lambda c: get_color(c, isinstance(new_options, rc.PrintHtmlOptions))  # type: ignore

        # Annotate mapped nodes with an example datum
        def get_data_comment(c: rc.Circuit) -> str:
            if scrubbed_matcher(c):
                return scrubbed_nodes[c].str_samplers(sampled_inputs=self.sampled_inputs)
            elif c in scrubbed_inputs:
                return str(self.sampled_inputs[scrubbed_inputs[c]])
            else:
                return ""

        new_options.commenters += [get_data_comment] if print_data else []

        # traversal
        if isinstance(new_options, rc.PrintOptions):
            new_options.bijection = False
        new_options.traversal = rc.restrict(new_options.traversal, term_early_at=term_at_matcher)

        if repr:
            assert isinstance(new_options, rc.PrintOptions)
            return self.circuit.repr(new_options)
        else:
            self.circuit.print(new_options)
